# -*- coding:utf-8 -*-

from typing import Dict
import os
import textwrap

from .base import ShellCompletion


class BashCompletion(ShellCompletion):

    _SHELL: str = "bash"

    try:
        _INJECT_PATH: str = os.environ["HOME"] + "/.bashrc"
    except:
        _INJECT_PATH: str = ""

    _template_source: str = textwrap.dedent(
        """\
        #!/usr/env bash

        %(func_name)s(){
          if [[ "${COMP_CWORD}" == "1" ]];then
              COMP_WORD="%(complete_vars)s"
              COMPREPLY=($(compgen -W "$COMP_WORD" -- ${COMP_WORDS[${COMP_CWORD}]}))
          fi
        }

        complete -F %(func_name)s %(prop)s
        """
    )

    def generate(self) -> str:
        # TODO:improve `bash` completion script.
        comp_keys = []

        _arguments, _positions, _sub_opts = self._parse(self.complete_vars["args"])

        for x in _arguments:
            comp_keys.extend(x[0].split())

        sub_q = [_sub_opts]
        while sub_q:
            temp: Dict = sub_q.pop(0)
            for opt_name, p in temp.items():
                comp_keys.append(opt_name)

                if _a := p.get("_arguments"):
                    for x in _a:
                        comp_keys.extend(x[0].split())
                if _opts := p.get("_sub_opts"):
                    sub_q.insert(-1, _opts)

        return " ".join(comp_keys)



# -*- coding:utf-8 -*-

from typing import Callable, Dict, Iterable, List, Optional, Tuple, ByteString, Union
import sys, subprocess, asyncio
from math import sqrt
from collections import Counter


def traceback_info(extra_msg: str = "null") -> str:
    """Get traceback infomation.

    Args:
        extra_msg (str, optional): extra custom message. Defaults to "null".

    Returns:
        str: formated traceback infomation.
    """

    exc_type, exc_value, exc_obj = sys.exc_info()
    if exc_type is None or exc_value is None or exc_obj is None:
        return ""

    err_value = exc_type.__name__
    lineno = exc_obj.tb_lineno
    filename = exc_obj.tb_frame.f_code.co_filename

    return (
        f"File {filename}, line {lineno}, {err_value}:{exc_value}, remark:[{extra_msg}]"
    )


def exec_cmd(
    *args, cwd: Optional[str] = None, reply: bool = True, decoding: bool = True
) -> Tuple[Union[str, ByteString, None], ...]:
    """Run system shell command.

    Args:
        reply (bool): whether return execute result. Default is True.
        decoding (bool): whether decode the return. Default is True.
    """
    _stderr: Optional[int] = subprocess.PIPE if reply else None
    _stdout: Optional[int] = subprocess.PIPE if reply else None

    _err: Union[str, ByteString, None] = None
    _rlt: Union[str, ByteString, None] = None

    try:
        # Take over the input stream and get the return information.
        with subprocess.Popen(
            " ".join(args), stderr=_stderr, stdout=_stdout, shell=True, cwd=cwd
        ) as proc:
            outres, errres = proc.communicate()
    except FileNotFoundError as e:
        _err = str(e).encode()
    else:
        _err, _rlt = errres, outres

    if not decoding:
        return _err, _rlt

    # decode
    if _err is not None:
        _err = _err.decode()
    if _rlt is not None:
        _rlt = _rlt.decode()
    return _err, _rlt


async def async_run_cmd(
    *args, cwd: Optional[str] = None, msg: Optional[str] = None, output: bool = True
):

    # receive (program, *args, ...), so must split the full cmd,
    # and unpack incoming.
    proc = await asyncio.create_subprocess_exec(
        *args,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        start_new_session=True,
        cwd=cwd,
    )
    res, err = await proc.communicate()

    if output:
        msg and print(msg)
        res and print(res.decode())
        err and print(err.decode())

    if proc.returncode != 0:
        return proc.returncode, args, cwd
    else:
        return proc.returncode, err.decode(), res.decode()


def exec_async_tasks(tasks: List[Callable]) -> List[str]:
    """Execute tasks asynchronously."""

    # The loop argument is deprecated since Python 3.8
    # and scheduled for removal in Python 3.10
    async def main():
        L = await asyncio.gather(*tasks)
        return L

    return asyncio.run(main())


def confirm(text: str = "", default: bool = True) -> bool:
    """Obtain confirmation results.
    Args:
        text (str): Confirmation prompt.
        default (bool): Result returned when unexpected input.

    Returns:
        (bool): Confirm result.
    """

    input_command: str = input(text).strip().lower()
    if input_command in {"n", "no", "N", "No"}:
        return False
    elif input_command in {"y", "yes", "Y", "Yes"}:
        return True
    else:
        return default


def similar_command(command: str, all_commands: Iterable) -> str:
    """Get the most similar command with K-NearestNeighbor.

    Args:
        command (str): command string.
        all_commands (list): The list of all command.

    Returns:
        (str): most similar command string.

    Docs test
        >>> commands = [
        ...     'branch', 'working tree', 'index', 'log', 'push',
        ...     'pull', 'tag','commit','conflict'
        ... ]
        >>> similar_command('br', commands)
        'branch'
        >>> similar_command('wo', commands)
        'working tree'
        >>> similar_command('com', commands)
        'commit'
    """

    #  The dictionary of letter frequency of all commands.
    words: dict = {word: dict(Counter(word)) for word in all_commands}
    # Letter frequency of command.
    fre = dict(Counter(command))
    # The distance between the frequency of each letter in the command
    # to be tested and all candidate commands, that is the difference
    # between the frequency of letters.
    frequency_difference: dict[str, list[int]] = {
        word: [fre[ch] - words[word].get(ch, 0) for ch in command]
        + [words[word][ch] - fre.get(ch, 0) for ch in word]
        for word in words
    }
    # Square of sum of squares of word frequency difference.
    frequency_sum_square: list[Tuple[str, int]] = list(
        map(
            lambda item: (item[0], int(sqrt(sum(map(lambda i: i ** 2, item[1]))))),
            frequency_difference.items(),
        )
    )

    # The value of `frequency_sum_square` is multiplied by the weight to find
    # the minimum.
    # Distance weight: compensate for the effect of length difference.
    # Compare Weight: The more similar the beginning, the higher the weight.
    # sourcery skip: inline-immediately-returned-variable, or-if-exp-identity
    min_frequency_command: str = min(
        frequency_sum_square,
        key=lambda item: item[1]
        * (
            len(command) / len(item[0])
            if len(command) / len(item[0])
            else len(item[0]) / len(command)
        )
        # Returns how many identical letters are compared from the head. sigmod to 0 ~ 1.
        * (1 / (len(list(filter(lambda i: i[0] == i[1], zip(command, item[0])))) + 1)),
    )[0]
    return min_frequency_command


# Mark the type corresponding to the file suffix.
# abcdefg hijklmn opq rst uvw xyz
SUFFIX_TYPE: Dict[str, str] = {
    "": "",
    "bat": "Batch",
    "c": "C",
    "cfg": "Properties",
    "conf": "Properties",
    "cpp": "C++",
    "cs": "C#",
    "css": "CSS",
    "dart": "Dart",
    "dea": "XML",
    "go": "Go",
    "gradle": "Groovy",
    "h": "C",
    "hpp": "C++",
    "htm": "HTML",
    "html": "HTML",
    "ini": "Ini",
    "java": "Java",
    "js": "Java Script",
    "json": "Json",
    "jsx": "React",
    "kt": "Kotlin",
    "launch": "XML",
    "less": "CSS",
    "lua": "Lua",
    "log": "Log",
    "m": "Object-C",
    "mm": "Object-C++",
    "markdown": "Markdown",
    "md": "Markdown",
    "msg": "ROS Message",
    "php": "PHP",
    "plist": "XML",
    "properties": "Propertie",
    "py": "Python",
    "r": "R",
    "rb": "Ruby",
    "rc": "Properties",
    "rs": "Rust",
    "rst": "reStructuredText",
    "ts": "Type Script",
    "tsx": "React",
    "sass": "CSS",
    "scss": "CSS",
    "sh": "Shell",
    "sql": "SQL",
    "srv": "ROS Message",
    "swift": "Swift",
    "toml": "Properties",
    "vb": "Visual Basic",
    "vim": "Vim Scirpt",
    "vue": "Vue",
    "xhtml": "HTML",
    "xml": "XML",
    "yaml": "YAML",
    "yml": "YAML",
    "zsh": "Shell",
}

# Mark the type of some special files.
SPECIAL_NAMES: Dict[str, str] = {
    "license": "LICENSE",
    "requirements.txt": "Pip requirement",
    "vimrc": "Vim Scirpt",
    "dockerfile": "Docker",
}


def adjudgment_type(file: str, original: bool = False) -> str:
    """Get file type.

    First, judge whether the file name is special, and then query the
    file suffix. Otherwise, the suffix or name will be returned as is.

    Args:
        file (str): file name string.
        original (bool, option): whether return origin when match failed.

    Returns:
        (str): file type.
    """

    if pre_type := SPECIAL_NAMES.get(file.lower(), None):
        return pre_type

    suffix = file.split(".")[-1]
    if suffix_type := SUFFIX_TYPE.get(suffix.lower(), None):
        return suffix_type
    else:
        return suffix if original else "unknown"


FILE_ICONS: Dict[str, str] = {
    "": "",
    "Batch": "",
    "C": "",
    "C#": "",
    "C++": "",
    "CSS": "",
    "Dart": "",
    "Groovy": "",
    "Go": "",
    "HTML": "",
    "Java": "",
    "Java Scirpt": "",
    "Lua": "",
    "Kotlin": "",
    "Markdown": "",
    "PHP": "",
    "Propertie": "",
    "Python": "",
    "R": "ﳒ",
    "React": "",
    "Ruby": "",
    "Rust": "",
    "ROS Message": "",
    "reStructuredText": "",
    "Shell": "",
    "Swift": "",
    "SQL": "",
    "snippets": "",
    "Type Scirpt": "",
    "Vim Scirpt": "",
    "Vue": "﵂",
    "YAML": "",
    "XML": "",
}


def get_file_icon(file_type: str) -> str:
    """According file type return icon.

    Args:
        file_type (str): type string.

    Returns:
        str: icon.
    """

    #     
    return FILE_ICONS.get(file_type, "")


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



# -*- coding:utf-8 -*-

from typing import Dict, List, Literal, Optional, Tuple, Union, Any
import os, re, json, stat
import logging
import threading
import concurrent.futures
from shutil import get_terminal_size

from plenty.table import Table
from plenty import get_console

from .common.utils import confirm, get_file_icon, adjudgment_type


Logger = logging.getLogger(__name__)


CounterFormatType = Literal["simple", "table"]


class CodeCounter(object):
    """Class of statistical code.

    Attributes:
        Absolute_Rules (dict): Precompiled rules.
        Suffix_Type (dict): Supported file suffix dictionary.
        Special_Name (dict): Type dict of special file name.
            subscript, and the codes of different levels are colored
            when the results are output.
        symbol (dict):
    """

    # The default rule is to count only files. Ignore all video, audio, fonts, binaries.
    Absolute_Rules: List[Dict] = [
        # Exclude `.git` folder.
        {"pattern": re.compile(r"\.git$|\.git\/"), "include": False},
        {
            # Exclude all picture formats.
            "pattern": re.compile(
                r"\.xbm$|\.tif$|\.pjp$|\.svgz$|\.jpg$|\.jpeg$|\.ico$|\.icns$|\.tiff$|\.gif$|\.svg$|\.jfif$|\.webp$|\.png$|\.bmp$|\.jpeg$|\.avif$",
                re.I,
            ),
            "include": False,
        },
        {
            # Exclude all video formats.
            "pattern": re.compile(
                r"\.avi$|\.rmvb$|\.rm$|\.asf$|\.divx$|\.mpg$|\.mpeg$|\.mpe$|\.wmv$|\.mp4$|\.mkv$|\.vob$",
                re.I,
            ),
            "include": False,
        },
        {
            # Exclude all audio frequency formats.
            "pattern": re.compile(
                r"\.mp3$|\.wma$|\.mid[i]?$|\.mpeg$|\.cda$|\.wav$|\.ape$|\.flac$|\.aiff$|\.au$",
                re.I,
            ),
            "include": False,
        },
        {
            # Exclude all font formats.
            "pattern": re.compile(
                r"\.otf$|\.woff$|\.woff2$|\.ttf$|\.eot$",
                re.I,
            ),
            "include": False,
        },
        {
            # Exclude some binary file.
            "pattern": re.compile(
                r"\.exe$|\.bin$",
                re.I,
            ),
            "include": False,
        },
    ]

    _Lock = threading.Lock()

    def __init__(
        self,
        *,
        count_path: Optional[str] = None,
        format_type: CounterFormatType = "table",
        use_ignore: bool = True,
        use_icon: bool = True,
        result_saved_path: str = "",
        whether_save: bool = True,
        color: bool = True,
        show_invalid: bool = False,
    ) -> None:
        """
        Args:
            count_path (str, optional): The path of needed count. Defaults to os.getcwd().
            use_ignore (bool, optional): Whether detect `.gitignore` file. Defaults to True.
            result_saved_path (str, optional): Result save path. Defaults to "".
            result_format (str, optional): Output format string. Defaults to "table".
            use_icon (bool, optional): Whether output with icon. Defaults to False.

        """
        # Store the rules obtained after processing.
        self.Rules: List = []

        self.count_path = count_path or os.getcwd()
        self.format_type = format_type
        self.use_ignore = use_ignore
        self.use_icon = use_icon
        self.result_saved_path = result_saved_path
        self.whether_save = whether_save
        self.color = color
        self.show_invalid = show_invalid

    def parse_gitignore2rule(self, root: str) -> None:
        """Process `.gitignore` files and add matching rules.

        Args:
            root (str): Absolute or relative path to the directory.
            files (list): The list of all file names under the `root` path.
        """

        root = root.replace("\\", "/")  # Unified symbol.
        ignore_path = os.path.join(root, ".gitignore")
        try:
            with open(ignore_path) as f:
                ignore_content = filter(
                    # Filter out comment lines.
                    lambda x: x and not x.startswith("#"),
                    map(
                        # Filter out white space lines.
                        # Replace `\` to `/` for windows.
                        lambda x: x.strip().replace("\\", "/"),
                        # Read the file and split the lines.
                        f.read().split("\n"),
                    ),
                )
        except PermissionError:
            if confirm(f"Can't read {ignore_path}, whether get jurisdiction[y/n]:"):
                os.chmod(ignore_path, stat.S_IXGRP)
                os.chmod(ignore_path, stat.S_IWGRP)
                self.parse_gitignore2rule(root)
        except Exception as e:
            print(f"Read gitignore error: {e}")
        else:
            for item in ignore_content:
                is_negative = item[0] == "!"
                if is_negative:
                    item = item[1:]

                slash_index = item.find("/")
                if slash_index == 0:
                    item = root + item
                elif slash_index in [-1, len(item) - 1]:
                    item = "/".join([root, "**", item])
                else:
                    item = "/".join([root, item])

                item = re.sub(r"([\{\}\(\)\+\.\^\$\|])", r"\1", item)  # escape char
                item = re.sub(r"(^|[^\\])\?", ".", item)
                item = re.sub(r"\/\*\*", "([\\\\/][^\\\\/]+)?", item)  # /**
                item = re.sub(r"\*\*\/", "([^\\\\/]+[\\\\/])?", item)  # **/
                item = re.sub(r"\*", "([^\\\\/]+)", item)  # for `*`
                item = re.sub(r"\?", "*", item)  # for `?``
                item = re.sub(r"([^\/])$", r"\1(([\\\\/].*)|$)", item)
                item = re.sub(r"\/$", "(([\\\\/].*)|$)", item)  # for trialing with `/`
                self.Rules.append({"pattern": re.compile(item), "include": is_negative})

    def matching_rules(self, full_path: str) -> bool:
        """Matching rules.

        Judge whether it is the required file according to the rule matching path.
        Returns `True` if the file not needs to be ignored, or `False` if needs.

        Args:
            full_path (str): File full path for matching.
        """

        # Precompiled rules have the highest priority.
        if list(
            filter(lambda rule: rule["pattern"].search(full_path), self.Absolute_Rules)
        ):
            return False

        # Matching the generated rules.
        res = list(filter(lambda rule: rule["pattern"].search(full_path), self.Rules))
        if not res:
            return True

        # If multiple rules match successfully, we think the last rule added has
        # the highest priority. Or if just one, this no problem also.
        return res[-1]["include"]
        # selected_rule = max(res, key=lambda rule: len(str(rule["pattern"])))

    def _sub_count(self, root: str, files: List) -> tuple:
        """Process handle use by `self.count`."""
        show_invailed = self.show_invalid

        result: dict[str[dict[str, int]]] = {}  # type: dict[str,dict]
        valid_counter = invalid_counter = 0
        invalid_list: list[str] = []
        total_size: int = 0

        for file in files:
            full_path = os.path.join(root, file)
            is_effective = self.matching_rules(full_path)
            if is_effective:
                try:
                    # Try read size of the valid file. Then do sum calc.
                    file_size = os.path.getsize(full_path)
                    total_size += file_size
                except:
                    Logger.warn(f"Can't read size of '{full_path}'")

                # Get file type.
                file_type = adjudgment_type(file, original=True)
                try:
                    with open(full_path) as f:
                        count = len(f.read().split("\n"))
                except Exception:
                    if show_invailed:
                        invalid_list.append(file)
                    invalid_counter += 1
                    continue
                else:
                    # Superposition.
                    if result.get(file_type) is None:
                        result[file_type] = {"files": 1, "lines": count}
                    else:
                        result[file_type]["files"] += 1
                        result[file_type]["lines"] += count
                    valid_counter += 1

        return result, total_size, valid_counter, invalid_counter, invalid_list

    @staticmethod
    def _walk_err_callback(e):
        """Handle of processing walk error."""
        print("Walk error: {0}".format(e))
        return

    def count(self) -> Tuple[Dict, List, List]:
        """Statistics file and returns the result dictionary for Python3.

        Args:
            root_path (str): The path is walk needed.
            use_ignore (bool): Whether ignore files in `.gitignore`. Defaults to True.
            progress (bool): Whether show processing. Defaults to True.

        Return:
            result (dict): Dictionary containing statistical results.
            invalid_list (list): invalid file list.
            total_size (int): the sum size of all valid files.

            result = {
                'Python': {
                    'files': 5,
                    'lines': 2124,
                }
            }
        """
        root_path = self.count_path
        use_ignore = self.use_ignore

        result = {}  # type: dict[str,dict]
        data_count = [0, 0, 0]  # [total_size, valid_count, invalid_count]
        invalid_list = []  # Invalid file list.

        def _callback(r):
            (
                _result,
                _total_size,
                _valid_counter,
                _invalid_counter,
                _invalid_list,
            ) = r.result()

            with self._Lock:
                for key, values in _result.items():
                    if result.get(key) is None:
                        result[key] = values
                    else:
                        result[key]["files"] += _result[key]["files"]
                        result[key]["lines"] += _result[key]["lines"]

                data_count[0] += _total_size
                data_count[1] += _valid_counter
                data_count[2] += _invalid_counter
                invalid_list.extend(_invalid_list)

        cpu: int = os.cpu_count() or 1
        max_queue: int = cpu * 200
        print("Detect CPU count: {0}, start record ...".format(cpu))

        with concurrent.futures.ProcessPoolExecutor() as pool:
            for root, _, files in os.walk(root_path, onerror=self._walk_err_callback):
                if not files:
                    continue

                # Judge whether the directory is valid. Invalid directories
                # do not traverse files.
                is_effective_dir = self.matching_rules(root)
                if not is_effective_dir:
                    continue

                # Process ``.gitignore`` file, add custom rules.
                if use_ignore and ".gitignore" in files:
                    self.parse_gitignore2rule(root)

                # If the number of files(exclude sub dir) in the directory is greater
                # than 15 and the number of enabled processes does not exceed the upper
                # limit, enable process traversal.
                if len(files) >= 15 and pool._queue_count < max_queue:
                    # Calling process.
                    future_result = pool.submit(self._sub_count, root, files)
                    future_result.add_done_callback(_callback)
                else:
                    (
                        _result,
                        _total_size,
                        _valid_counter,
                        _invalid_counter,
                        _invalid_list,
                    ) = self._sub_count(root, files)

                    with self._Lock:
                        for key, values in _result.items():
                            if result.get(key, None) is None:
                                result[key] = values
                            else:
                                result[key]["files"] += _result[key]["files"]
                                result[key]["lines"] += _result[key]["lines"]

                        data_count[0] += _total_size
                        data_count[1] += _valid_counter
                        data_count[2] += _invalid_counter
                        invalid_list.extend(_invalid_list)

        # Some processes may not be finished after synchronization.
        print("\nPlease wait calculate ...")

        return result, invalid_list, data_count[0]

    def _get_saved_path(self, root_path: str) -> str:
        file_name: str = (
            root_path.replace("/", "_").replace("\\", "_").replace(".", "_")
        )
        return os.path.join(self.result_saved_path, file_name)

    def load_recorded_result(self, root_path: str) -> Optional[Dict]:
        """Load count result."""
        file_path = self._get_saved_path(root_path)
        try:
            with open(file_path, "r") as rf:
                res = json.load(rf)
        except Exception:
            return None
        else:
            return res

    def save_result(self, result: Dict, root_path: str) -> bool:
        """Save count result.

        Generate name according to `root_path`, then try save the record
        result to [`TOOLS_HOME`/Counter].

        Args:
            result (dict): Statistical results.
            root_path (str): Traversal directory.

        Return:
            (bool): Whether saving successful.
        """

        file_path = self._get_saved_path(root_path)
        try:
            with open(file_path, "w" if os.path.isfile(file_path) else "x") as wf:
                json.dump(result, wf, indent=2)
        except Exception:
            return False
        else:
            return True

    def simple_output(self, result: Dict) -> str:
        gen = []
        for key, value in result.items():
            line = f"::{key}  (files:{value['files']:,} | lines:{value['lines']:,})"
            gen.append(line)

        return "\n".join(gen)

    def table_output(self, new: Dict, old: Dict) -> Table:
        use_icon = self.use_icon
        color = self.color

        tb = Table(title="[Code Counter Result]", title_style="bold")
        tb.add_column("Language")
        tb.add_column("Files")
        tb.add_column("Code lines")

        # Diff
        sum_lines = 0
        for key, value in new.items():
            files = value["files"]
            lines = value["lines"]
            # Calc sum code line.
            sum_lines += lines

            file_type_str = (
                "{0} {1}".format(get_file_icon(key), key) if use_icon else key
            )
            files_str = f"{files:,}"
            lines_str = f"{lines:,}"

            if color:
                file_type_str = f"`{file_type_str}`<cyan>"
                files_str = f"`{files_str}`<{self.color_index(files)}>"
                lines_str = f"`{lines_str}`<{self.color_index(lines)}>"

            # Compare change.
            if isinstance(old, dict) and old.get(key) is not None:
                old_files = old.get(key).get("files", None)
                old_lines = old.get(key).get("lines", None)

                if old_files and old_files != files:
                    files_change = "{:+}".format(files - old_files)
                else:
                    files_change = ""

                if old_lines and old_lines != lines:
                    lines_change = "{:+}".format(lines - old_lines)
                else:
                    lines_change = ""

            else:
                files_change = lines_change = ""

            if color:
                files_change_str = f"`{files_change}`<{'#98fb98' if files_change.startswith('+') else '#ff6347'}>"
                lines_change_str = f"`{lines_change}`<{'#98fb98' if lines_change.startswith('+') else '#ff6347'}>"
            else:
                files_change_str = files_change
                lines_change_str = lines_change

            tb.add_row(
                file_type_str,
                f"{files_str} {files_change_str}",
                f"{lines_str} {lines_change_str}",
            )

        # Print total and change graph.
        tb.caption = " Total: {0} lines".format(sum_lines)
        return tb

    @classmethod
    def color_index(
        cls, count: int, level_color: Union[List, Tuple, None] = None
    ) -> int:
        # Colors displayed for different code quantities.
        if level_color is None:
            level_color = (
                "green",
                "#EBCB8C",  # yellow
                "#FF6347",  # tomato
                "#C71585",  # middle violet red
                "#87CEFA",  # skyblue
            )
        index = len(str(count // 1000))
        return level_color[-1] if index > len(level_color) else level_color[index - 1]

    def generate_format_output(self, new: Dict, old: Optional[Dict] = None) -> Any:
        """Print result with color and diff.

        If the console width is not enough, the output is simple.

        Args:
            new (dict): Current statistical results.
            old (dict|None): The results saved in the past may not exist.
        """

        result_format = self.format_type

        needed_width: int = 67
        width, _ = get_terminal_size()
        if result_format == "simple" or width < needed_width:
            return self.simple_output(new)

        elif result_format == "table":
            return self.table_output(new, old)

    def run(
        self,
        whether_output: bool = True,
        *,
        count_path: Optional[str] = None,
        format_type: Optional[CounterFormatType] = None,
        use_ignore: Optional[bool] = None,
        use_icon: Optional[bool] = None,
        result_saved_path: Optional[str] = None,
        whether_save: Optional[bool] = None,
        color: Optional[bool] = None,
        show_invalid: Optional[bool] = None,
    ) -> None:
        if count_path is not None:
            self.count_path = count_path
        if format_type is not None:
            self.format_type = format_type
        if use_ignore is not None:
            self.use_ignore = use_ignore
        if use_icon is not None:
            self.use_icon = use_icon
        if result_saved_path is not None:
            self.result_saved_path = result_saved_path
        if whether_save is not None:
            self.whether_save = whether_save
        if color is not None:
            self.color = color
        if show_invalid is not None:
            self.show_invalid = show_invalid

        result, invalid_list, total_size = self.count()

        old_result = self.load_recorded_result(self.count_path)

        # diff print.
        if whether_output:
            output = self.generate_format_output(result, old_result)
            get_console().echo(output)

            if self.whether_save:
                self.save_result(result, self.count_path)
            if (
                self.show_invalid
                and invalid_list
                and confirm("Whether print invalid file list?[y/n]", default=False)
            ):
                print(invalid_list)

            # optimize size unit.
            size_unit = ["byte", "KB", "MB", "GB"]
            for i in range(3):
                if total_size >= 1024:
                    total_size /= 1024
                else:
                    break
            else:
                i = 3
            print("Files total size: {0:.2f}{1}".format(total_size, size_unit[i]))

        return result



from typing import Any, Callable
from functools import wraps
import time, contextlib, inspect


def time_it(fn: Callable) -> Callable:
    """Print the overall running time.
    When recursive calls exist, only the outermost layer is printed.
    """
    time_it.deep = 0  # Mark recursion levels.
    time_unit = ["second", "mintue", "hour"]

    @wraps(fn)
    def wrap(*args, **kwargs):
        time_it.deep += 1
        start_time = time.time()
        res = None
        with contextlib.suppress(SystemExit, EOFError):
            res = fn(*args, **kwargs)
        time_it.deep -= 1

        # Indicates that the decorated method does not or end a recursive call.
        if time_it.deep == 0:
            used_time = time.time() - start_time

            # Do unit optimization.
            for i in range(2):
                if used_time >= 60:
                    used_time /= 60
                else:
                    break
            else:
                i = 2
            print("\nruntime: {0:.2f} {1}".format(used_time, time_unit[i]))
        return res

    return wrap


def dynamic_default_attrs(fn: Callable, **kwds: Any) -> Callable:
    """Set default parameters dynamically.

    Receive a method and several named parameters, which will be saved.
    When the method is called, it will be filled in automatically, and
    the incoming value will replace the dynamic default value. It can be
    used as a decorator.
    """

    _kwds = kwds
    _fn_params = list(inspect.signature(fn).parameters.keys())
    _fn_params_len = len(_fn_params)

    @wraps(fn)
    def wrap(*args, **kwargs) -> Any:
        final_kwargs = {**_kwds, **kwargs}
        args_len = len(args)

        if extra_len := (args_len + len(final_kwargs) - _fn_params_len) > 0:
            used_list = _fn_params[args_len - extra_len : args_len]
            for used in used_list:
                final_kwargs.pop(used)

        return fn(*args, **final_kwargs)

    return wrap



from functools import wraps
from pyinstrument import Profiler


def analyze_it(fn):
    """Analysis of decorated methods."""

    @wraps(fn)
    def inner(*args, **kwargs):
        profiler = Profiler()

        with profiler:
            res = fn(*args, **kwargs)
        profiler.print()
        return res

    return inner



# -*- coding:utf-8 -*-

from typing import Callable, Dict, List, Optional, Tuple, Union
import os, re, textwrap, random, logging

from plenty import get_console
from plenty.str_utils import shorten

from ..common.utils import exec_cmd, confirm, similar_command, traceback_info
from ..common.singleton import Singleton
from .shortcmds import GIT_CMDS, CommandType

Log = logging.getLogger(__name__)


def get_extra_cmds(name: str, path: str) -> Dict:
    """Get custom cmds.

    Load the `extra_cmds.py` file under PIGIT HOME, check whether `extra_cmds`
    exists, and return it. If not have, return a empty dict.

    Returns:
        (dict[str,str]): extra cmds dict.
    """
    import importlib.util

    extra_cmds = {}

    if os.path.isfile(path):
        try:
            # load a module form localtion.
            spec = importlib.util.spec_from_file_location(name, path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        except Exception:
            Log.error(traceback_info(f"Can't load file '{path}'."))
        else:
            try:
                extra_cmds = module.extra_cmds  # type: ignore
            except AttributeError:
                Log.error("Can't found dict name is 'extra_cmds'.")

    return extra_cmds


class ShortGitter(metaclass=Singleton):
    """Git short command processor."""

    def __init__(
        self,
        extra_cmds: Optional[dict] = None,
        command_prompt: bool = True,
        show_original: bool = True,
        **kwargs,
    ) -> None:

        self.use_recommend = command_prompt
        self.show_original = show_original

        # Init commands.
        self.cmds = GIT_CMDS
        if extra_cmds:
            if not isinstance(extra_cmds, dict):
                raise TypeError("Custom cmds must be a dict.") from None
            self.cmds.update(extra_cmds)

    @staticmethod
    def color_command(command: str) -> str:
        """Return the command string with color.
        prog      : green;
        command   : yellow;
        arguments : skyblue;
        values    : white.
        """

        handle = re.match(r"(\w+)\s+(\w+)", command)
        prop = handle[1]
        cmd = handle[2]
        next_position = handle.span()[1]

        color_command = f"b`{prop}`<ok> b`{cmd}`<goldenrod> "

        arguments = re.findall(
            r"\s+(-?-?[^\s]+)(=?([\w-]+:)?(\".*?\"|\'.*?\'))?", command[next_position:]
        )
        for arg_handle, value, _, _ in arguments:
            color_command += f"`{arg_handle}`<sky_blue>{value} "

        return color_command

    def process_command(
        self, command_string: str, args: Optional[Union[List, Tuple]] = None
    ) -> Tuple[int, str]:
        """Process command and arguments.

        Args:
            command_ (str): short command string
            args (list|None, optional): command arguments. Defaults to None.
        """

        option: Optional[Dict[str, Dict]] = self.cmds.get(command_string, None)

        # Invalid, if need suggest.
        if option is None:
            return (
                1,
                f"Don't support this command: `{command_string}`<error>, "
                "please try `--show-commands`<gold>",
            )

        command: Optional[Union[str, Callable]] = option.get("command")
        # Has no command can be executed.
        if not command:
            return 2, "`Invalid custom short command, nothing can to exec.`<error>"

        if not option.get("has_arguments", False) and args:
            get_console().echo(
                f"`The command does not accept parameters. Discard {args}.`<error>"
            )
            args = []

        if isinstance(command, Callable):
            try:
                command(args)
            except Exception as e:
                return 3, f"`{e}`<error>"
        elif isinstance(command, str):
            if args:
                command = " ".join([command, *args])
            if self.show_original:
                get_console().echo(f":rainbow:  {self.color_command(command)}")
            exec_cmd(command, reply=False)
        else:
            return 5, "`The type of command not supported.`<error>"

        return 0, ""

    def do(self, command_string: str, args: Optional[Union[List, Tuple]] = None):
        """Process command and arguments."""

        code, msg = self.process_command(command_string, args)
        if code == 0:
            pass
        elif code == 1 and self.use_recommend:  # check config.
            predicted_command = similar_command(command_string, self.cmds.keys())
            if confirm(
                get_console().render_str(
                    f":thinking: The wanted command is `{predicted_command}`<ok> ?[y/n]:"
                )
            ):
                self.do(predicted_command, args=args)
        else:
            get_console().echo(msg)

    # ============================
    # Print command help message.
    # ============================
    def _generate_help_by_key(
        self, _key: str, use_color: bool = True, max_width=90
    ) -> str:
        """Generate one help by given key.

        Args:
            _key (str): Short command string.
            use_color (bool, optional): Whether color help message. Defaults to True.

        Returns:
            (str): Help message of one command.
        """

        help_position = 15
        msg_max_width = max_width - help_position

        # Get help message and command.
        _help: str = self.cmds[_key].get("help", "").strip()
        if _help:
            _help = textwrap.wrap(_help, msg_max_width)
            help_msg = _help[0] + "\n"
            for line in _help[1:]:
                help_msg += "%*s%s\n" % (help_position, "", line)
        else:
            help_msg = ""

        _command = self.cmds[_key].get("command", "ERROR: empty command.")
        if callable(_command):
            _command = f"Func: {_command.__name__}"

        _command = shorten(_command, msg_max_width, placeholder="...")
        command_msg = "%*s%s" % (help_position, "", _command) if help_msg else _command

        if use_color:
            return f"  `{_key:<13}`<ok>{help_msg}`{command_msg}`<gold>"
        else:
            return f"  {_key:<12} {_help}{command_msg}"

    def print_help(self) -> None:
        """Print help message."""
        print("These are short commands that can replace git operations:")
        for key in self.cmds.keys():
            msg = self._generate_help_by_key(key)
            get_console().echo(msg)

    def print_help_by_type(self, command_type: str) -> None:
        """Print a part of help message.

        Print the help information of the corresponding part according to the
        incoming command type string. If there is no print error prompt for the
        type.

        Args:
            command_type (str): A command type of `TYPE`.
        """

        # Process received type.
        command_type = command_type.capitalize().strip()

        # Checking the type whether right.
        if command_type not in CommandType.__members__:
            get_console().echo(
                "`There is no such type.`<error>\n"
                "Please use `git --types`<ok> to view the supported types."
            )

            if self.use_recommend:
                predicted_type = similar_command(
                    command_type, CommandType.__members__.keys()
                )
                if confirm(
                    get_console().render_str(
                        f":thinking: The wanted type is `{predicted_type}`<ok> ?[y/n]:"
                    )
                ):
                    self.print_help_by_type(predicted_type)
            return None

        # Print help.
        print("These are the orders of {0}".format(command_type))
        for k, v in self.cmds.items():
            belong = v.get("belong", CommandType.Extra)
            # Prevent the `belong` attribute from being set in the custom command.
            if isinstance(belong, CommandType) and belong.value == command_type:
                msg = self._generate_help_by_key(k)
                get_console().echo(msg)

    @classmethod
    def print_types(cls) -> None:
        """Print all command types with random color."""
        res = []

        for member in CommandType:
            color_str = "#{:02X}{:02X}{:02X}".format(
                random.randint(70, 255),
                random.randint(70, 255),
                random.randint(70, 255),
            )

            res.append(f"`{member.value}`<{color_str}>")

        get_console().echo(" ".join(res))



# -*- coding:utf-8 -*-

"""Command methods.

This file encapsulates some methods corresponding to command.
All methods are must and only accept an `args` parameter -- a list or tuple
of parameters to be processed. If don't need will get a empty list or tuple.

Like this:
>>> def func(args: Union[list, tuple]) -> None:
>>>     # your code.
>>>     pass
"""

from typing import List, Tuple, Union

from plenty import echo

from ..common.utils import exec_cmd


def add(args: Union[List, Tuple]) -> None:
    """git add option.

    Args:
        args (list): arguments list, maybe empty. need file string.
    """

    args_str = " ".join(args) if args else " ."
    echo(
        ":rainbow: Storage file: {0}".format(
            "all" if args_str.strip() == "." else args_str
        )
    )
    exec_cmd(f"git add {args_str}", reply=False)


def fetch_remote_branch(args: Union[List, Tuple]) -> None:
    "Fetch a remote branch to local and with the same name."

    branch = args[0] if len(args) > 1 else None

    if branch:
        exec_cmd("git fetch origin {0}:{0} ".format(branch), reply=False)
    else:
        echo("`This option need a branch name.`<error>")


def set_email_and_username(args: Union[List, Tuple]) -> None:
    """Set git username and email with interaction.

    Args:
        args (list): arguments list, maybe empty.
    """

    print("Set the interactive environment of user name and email ...")

    is_global = next(
        (" --global " for i in args if i.strip() in ["-g", "--global"]), ""
    )

    if is_global:
        print("Now set for global.")
    else:
        print("Now set for local.")

    while True:
        name = input("Please input username:").strip()

        if name:
            break
        else:
            echo("`Name is empty.`<error>")

    import re

    email_re = re.compile(r"^[0-9a-zA-Z_]{0,19}@[0-9a-zA-Z]{1,13}\.[com,cn,net]{1,3}$")
    while True:
        email = input("Please input email:")

        if email_re.match(email) is not None:
            break
        else:
            echo("`Bad mailbox format.`<error>")

    if exec_cmd(f"git config user.name {name} {is_global}", reply=False) and exec_cmd(
        f"git config user.email {email} {is_global}", reply=False
    ):
        echo("`Successfully set.`<ok>")
    else:
        echo("`Failed. Please check log.`<error>")



# -*- coding:utf-8 -*-
import os, sys

# Add source environment.
_PIGIT_PATH = os.path.dirname(os.path.dirname(__file__))
sys.path.insert(0, _PIGIT_PATH)

TEST_PATH = os.path.dirname(__file__)

# /opt/homebrew/opt/python@3.8/bin/python3 -m pytest



# -*- coding:utf-8 -*-

from typing import Dict, List
import os, re, textwrap, logging
from distutils.util import strtobool

from plenty.style import Color

from .common.utils import confirm, traceback_info
from .common.singleton import Singleton

Logger = logging.getLogger(__name__)

CONF_ERROR = "==error=="


class ConfigError(Exception):
    """Config error. Using by `Config`."""

    pass


class Config(metaclass=Singleton):
    """PIGIT configuration class."""

    CONFIG_TEMPLATE: str = textwrap.dedent(
        """\
        #? Config file for pigit v. {version}

        #  ____ ___ ____ ___ _____                            __ _
        # |  _ \\_ _/ ___|_ _|_   _|           ___ ___  _ __  / _(_) __ _
        # | |_) | | |  _ | |  | |_____ _____ / __/ _ \\| '_ \\| |_| |/ _` |
        # |  __/| | |_| || |  | |_____|_____| (_| (_) | | | |  _| | (_| |
        # |_|  |___\\____|___| |_|            \\___\\___/|_| |_|_| |_|\\__, |
        #                                     {version:>20} |___/
        # Git-tools -- pigit configuration.

        # (bool) Show original git command.
        cmd_show_original={cmd_show_original}

        # (bool) Is it recommended to correct when entering wrong commands.
        cmd_recommend={cmd_recommend}

        # (float) Display time of help information in interactive mode.
        tui_help_showtime={tui_help_showtime}

        # (bool) Whether to use the ignore configuration of the `.gitignore` file.
        counter_use_gitignore={counter_use_gitignore}

        # (bool) Whether show files that cannot be counted.
        counter_show_invalid={counter_show_invalid}

        # (bool) Whether show files icons. Font support required, like: 'Nerd Font'
        counter_show_icon={counter_show_icon}

        # Output format of statistical results. Supported: [table, simple]
        # When the command line width is not enough, the `simple ` format is forced.
        counter_format={counter_format}

        # Git local config print format. Supported: [table, normal]
        git_config_format={git_config_format}

        # Control which parts need to be displayed when viewing git repository information.
        # Support: (path,remote,branch,log,summary)
        repo_info_include={repo_info_include}

        # (bool) Whether auto append path to repos.
        repo_auto_append={repo_auto_append}

        # (bool) Whether run PIGIT in debug mode.
        debug_open={debug_open}

        # (bool) Whether output log in terminal.
        log_output={log_output}

        """
    )

    _keys: List[str] = [
        "cmd_show_original",
        "cmd_recommend",
        "tui_help_showtime",
        "counter_use_gitignore",
        "counter_show_invalid",
        "counter_show_icon",
        "counter_format",
        "git_config_format",
        "repo_info_include",
        "repo_auto_append",
        "debug_open",
        "log_output",
    ]

    # ======================
    # config default values.
    # ======================

    # cmd processor conf
    cmd_show_original: bool = True
    cmd_recommend: bool = True

    # tui conf
    tui_help_showtime: float = 1.5

    # code counter conf
    counter_use_gitignore: bool = True
    counter_show_invalid: bool = False
    counter_show_icon: bool = False
    counter_format: str = "table"  # table, simple
    _supported_result_format: List = ["table", "simple"]

    # info conf
    git_config_format: str = "table"
    _supported_git_config_format: List = ["normal", "table"]

    repo_info_include: List[str] = ["remote", "branch", "log"]

    # repo conf
    repo_auto_append: bool = False

    # setting conf
    debug_open: bool = False
    log_output: bool = False

    # Store warning messages.
    warnings: List = []

    def __init__(
        self, path: str, version: str = "unknown", auto_load: bool = True
    ) -> None:
        self.config_file_path: str = path
        self.current_version: str = version
        self.conf: Dict = {}

        if auto_load:
            self.load_config()

    def output_warnings(self) -> None:
        """Output config warning info and return self object.

        Returns:
            self (Config): single `Config` object.
        """

        if self.warnings:
            print("Config Warning Info")
            print("=" * 30)
            for warning in self.warnings:
                print(warning)
            print("=" * 30)
            self.warnings = []

        return self

    def check_and_set_value(self, key: str, value: str, new_config: Dict) -> None:
        if key not in self._keys:
            self.warnings.append(f"'{key}' is not be supported!")
        elif type(getattr(self, key)) == int:
            try:
                new_config[key] = int(value)
            except ValueError:
                self.warnings.append(f'Config key "{key}" should be an integer!')
        elif type(getattr(self, key)) == float:
            try:
                new_config[key] = float(value)
            except ValueError:
                self.warnings.append(f'Config key "{key}" should be a float!')
        elif type(getattr(self, key)) == bool:
            try:
                # True values are y, yes, t, true, on and 1;
                # false values are n, no, f, false, off and 0.
                # Raises ValueError if val is anything else.
                new_config[key] = bool(strtobool(value))
            except ValueError:
                self.warnings.append(f'Config key "{key}" can only be True or False!')
        elif type(getattr(self, key)) == str:
            if "color" in key and not Color.is_color(value):
                self.warnings.append(
                    f'Config key "{key}" should be RGB string, like: "#FF0000".'
                )
            else:
                new_config[key] = value
        elif type(getattr(self, key)) == list:
            if re.match(r"^\[[\w\s0-9\'\"_]+,?([\s\w0-9\'\"_]+,?)*\]$", value):
                new_config[key] = eval(value)
            else:
                self.warnings.append(f'Config key "{key}" not support `{value}`.')

    def read_config(self) -> None:
        new_config = self.conf
        config_file = self.config_file_path

        with open(config_file) as cf:
            for line in cf:
                line = line.strip()
                if line.startswith("#? Config"):
                    new_config["version"] = line[line.find("v. ") + 3 :]
                    continue
                if line.startswith("#"):
                    # comment line.
                    continue
                if "=" not in line:
                    # invalid line.
                    continue

                # remove line comment.
                if line_comment_idx := line.find("#") > 0:
                    line = line[:line_comment_idx]

                # processing.
                key_string, value_string = line.split("=", maxsplit=1)
                key_string = key_string.strip()
                value_string = value_string.strip().strip('"')

                # checking.
                self.check_and_set_value(key_string, value_string, new_config)

    def parse_conf(self) -> None:
        new_config = self.conf

        if (  # check codecounter output format whether supported.
            "codecounter_result_format" in new_config
            and new_config["codecounter_result_format"]
            not in self._supported_result_format
        ):
            new_config["codecounter_result_format"] = CONF_ERROR
            self.warnings.append(
                'Config key "{0}" support must in {1}'.format(
                    "codecounter_result_format", self._supported_result_format
                )
            )

        if (
            "git_config_format" in new_config
            and new_config["git_config_format"] not in self._supported_git_config_format
        ):
            new_config["git_config_format"] = CONF_ERROR
            self.warnings.append(
                'Config key "{0}" support must in {1}'.format(
                    "git_config_format", self._supported_git_config_format
                )
            )

        version = self.current_version
        if "version" in new_config and not (
            # If unknown current version or
            # If the version is right or
            # If current version is a [beta, alpha, dev] verstion then will not tip.
            # Else if version is not right will tip.
            version == "unknown"
            or new_config["version"] == version
            or "beta" in version
            or "alpha" in version
            or "dev" in version
        ):
            print(new_config["version"])
            self.warnings.append(
                "The current configuration file is not up-to-date."
                "You'd better recreate it."
            )

    def load_config(self) -> None:
        try:
            self.read_config()
        except Exception:
            Logger.error(traceback_info())
            self.warnings.append(
                f"Can not load the config file. Path: {self.config_file_path}"
            )

        self.parse_conf()

        # setting config.
        for key in self._keys:
            if key in self.conf.keys() and self.conf[key] != CONF_ERROR:
                setattr(self, key, self.conf[key])
            else:
                self.conf[key] = getattr(self, key)

    def create_config_template(self) -> None:
        parent_dir = os.path.dirname(self.config_file_path)
        if not os.path.isdir(parent_dir):
            os.makedirs(parent_dir, exist_ok=True)

        if os.path.exists(self.config_file_path) and not confirm(
            "Configuration exists, overwrite? [y/n]"
        ):
            return None

        # Try to load already has config.
        self.load_config()
        self.conf["version"] = self.current_version

        # Write config with already exist custom settings.
        try:
            with open(
                self.config_file_path,
                "w" if os.path.isfile(self.config_file_path) else "x",
            ) as f:
                f.write(self.CONFIG_TEMPLATE.format(**self.conf))
        except Exception:
            Logger.error(traceback_info())
            print("Fail to create config.")
        else:
            print("Successful.")



# -*- coding:utf-8 -*-

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .screen import Screen


class ExitEventLoop(Exception):
    """Get to exit current event loop."""


class EventLoop(object):
    """
    Params:
        screen (Screen): screen to use, default is a new :class:`.screen.Screen`.
        input_handle (Input): input handle to use, default is a new :class:`.input:PosixInput`.
        real_time (bool): whether refresh screen real time.
    """

    def __init__(
        self,
        screen: "Screen" = None,
        input_handle=None,
        real_time: bool = False,
        debug: bool = False,
    ):
        self._real_time = real_time
        self.debug = debug

        # Init screen object.
        if not screen:
            from .screen import Screen

            screen = Screen()
        self._screen = screen

        # Init keyboard handle object.
        if not input_handle:
            # XXX: now not support windows.
            from .input import PosixInput, is_mouse_event

            input_handle = PosixInput()
            # adjust the input whether is a mouse event.
            self.is_mouse_event = is_mouse_event
        self._input_handle = input_handle

    def set_input_timeouts(self, timeout: float):
        self._input_handle.set_input_timeouts(timeout)

    def _loop(self):
        """Main loop"""
        if input_key := self._input_handle.get_input():
            first_one = input_key[0]

            if first_one == "window resize":
                self._screen.resize()
            elif hasattr(self, "is_mouse_event") and self.is_mouse_event(first_one):
                self._screen.process_mouse(first_one)
            else:
                self._screen.process_input(first_one)
        elif self._real_time:
            self._screen.render()

    def _run(self):
        try:
            while True:
                self._loop()
        except (ExitEventLoop, KeyboardInterrupt, EOFError):
            self._input_handle.stop()

    def run(self):
        with self._screen:
            self._input_handle.start()
            self._run()



# -*- coding:utf-8 -*-
import os, sys
import tty, termios, select, fcntl
import signal
from subprocess import Popen, PIPE


# ===========
# Predefined
# ===========
ord2 = lambda x: x
B = lambda x: x.encode("iso8859-1")  # noqa: E731

# ==================
# Terminal encoding
# ==================
_byte_encoding = None


def set_byte_encoding(enc):
    assert enc in ("utf8", "narrow", "wide")
    global _byte_encoding
    _byte_encoding = enc


def get_byte_encoding():
    return _byte_encoding


def detect_encoding():
    # Try to determine if using a supported double-byte encoding
    import locale

    initial = locale.getlocale()
    try:
        try:
            locale.setlocale(locale.LC_ALL, "")
        except locale.Error:
            pass
        return locale.getlocale()[1] or ""
    except ValueError as e:
        # with invalid LANG value python will throw ValueError
        if e.args and e.args[0].startswith("unknown locale"):
            return ""
        else:
            raise
    finally:
        try:
            locale.setlocale(locale.LC_ALL, initial)
        except locale.Error:
            pass


if "detected_encoding" not in locals():
    detected_encoding = detect_encoding()
else:
    assert 0, "It worked!"

_target_encoding = None
_use_dec_special = True


def set_encoding(encoding):
    """
    Set the byte encoding to assume when processing strings and the
    encoding to use when converting unicode strings.
    """
    encoding = encoding.lower()

    global _target_encoding, _use_dec_special

    if encoding in ("utf-8", "utf8", "utf"):
        set_byte_encoding("utf8")

        _use_dec_special = False
    elif encoding in (
        "euc-jp",  # JISX 0208 only
        "euc-kr",
        "euc-cn",
        "euc-tw",  # CNS 11643 plain 1 only
        "gb2312",
        "gbk",
        "big5",
        "cn-gb",
        "uhc"
        # these shouldn't happen, should they?
        ,
        "eucjp",
        "euckr",
        "euccn",
        "euctw",
        "cncb",
    ):
        set_byte_encoding("wide")

        _use_dec_special = True
    else:
        set_byte_encoding("narrow")
        _use_dec_special = True

    # if encoding is valid for conversion from unicode, remember it
    _target_encoding = "ascii"
    try:
        if encoding:
            u"".encode(encoding)
            _target_encoding = encoding
    except LookupError:
        pass


# ==============
# Util function
# ==============
def within_double_byte(text, line_start, pos):
    """Return whether pos is within a double-byte encoded character.
    text -- byte string in question
    line_start -- offset of beginning of line (< pos)
    pos -- offset in question
    Return values:
    0 -- not within dbe char, or double_byte_encoding == False
    1 -- pos is on the 1st half of a dbe char
    2 -- pos is on the 2nd half of a dbe char
    """
    assert isinstance(text, bytes)
    v = ord2(text[pos])

    if v >= 0x40 and v < 0x7F:
        # might be second half of big5, uhc or gbk encoding
        if pos == line_start:
            return 0

        if ord2(text[pos - 1]) >= 0x81:
            if within_double_byte(text, line_start, pos - 1) == 1:
                return 2
        return 0

    if v < 0x80:
        return 0

    i = pos - 1
    while i >= line_start:
        if ord2(text[i]) < 0x80:
            break
        i -= 1

    if (pos - i) & 1:
        return 1
    return 2


def is_mouse_event(ev):
    return type(ev) == tuple and len(ev) == 4 and ev[0].find("mouse") >= 0


def is_mouse_press(ev):
    return ev.find("press") >= 0


# =================
## Input sequences
# =================
class MoreInputRequired(Exception):
    pass


def escape_modifier(digit):
    mode = ord(digit) - ord("1")
    return (
        "shift " * (mode & 1)
        + "meta " * ((mode & 2) // 2)
        + "ctrl " * ((mode & 4) // 4)
    )


# yapf: disable
input_sequences = [
    ('[A','up'),('[B','down'),('[C','right'),('[D','left'),
    ('[E','5'),('[F','end'),('[G','5'),('[H','home'),

    ('[1~','home'),('[2~','insert'),('[3~','delete'),('[4~','end'),
    ('[5~','page up'),('[6~','page down'),
    ('[7~','home'),('[8~','end'),

    ('[[A','f1'),('[[B','f2'),('[[C','f3'),('[[D','f4'),('[[E','f5'),

    ('[11~','f1'),('[12~','f2'),('[13~','f3'),('[14~','f4'),
    ('[15~','f5'),('[17~','f6'),('[18~','f7'),('[19~','f8'),
    ('[20~','f9'),('[21~','f10'),('[23~','f11'),('[24~','f12'),
    ('[25~','f13'),('[26~','f14'),('[28~','f15'),('[29~','f16'),
    ('[31~','f17'),('[32~','f18'),('[33~','f19'),('[34~','f20'),

    ('OA','up'),('OB','down'),('OC','right'),('OD','left'),
    ('OH','home'),('OF','end'),
    ('OP','f1'),('OQ','f2'),('OR','f3'),('OS','f4'),
    ('Oo','/'),('Oj','*'),('Om','-'),('Ok','+'),

    ('[Z','shift tab'),
    ('On', '.'),

    ('[200~', 'begin paste'), ('[201~', 'end paste'),
] + [
    (prefix + letter, modifier + key)
    for prefix, modifier in zip('O[', ('meta ', 'shift '))
    for letter, key in zip('abcd', ('up', 'down', 'right', 'left'))
] + [
    ("[" + digit + symbol, modifier + key)
    for modifier, symbol in zip(('shift ', 'meta '), '$^')
    for digit, key in zip('235678',
        ('insert', 'delete', 'page up', 'page down', 'home', 'end'))
] + [
    ('O' + chr(ord('p')+n), str(n)) for n in range(10)
] + [
    # modified cursor keys + home, end, 5 -- [#X and [1;#X forms
    (prefix+digit+letter, escape_modifier(digit) + key)
    for prefix in ("[", "[1;")
    for digit in "12345678"
    for letter,key in zip("ABCDEFGH",
        ('up','down','right','left','5','end','5','home'))
] + [
    # modified F1-F4 keys -- O#X form
    ("O"+digit+letter, escape_modifier(digit) + key)
    for digit in "12345678"
    for letter,key in zip("PQRS",('f1','f2','f3','f4'))
] + [
    # modified F1-F13 keys -- [XX;#~ form
    ("["+str(num)+";"+digit+"~", escape_modifier(digit) + key)
    for digit in "12345678"
    for num,key in zip(
        (3,5,6,11,12,13,14,15,17,18,19,20,21,23,24,25,26,28,29,31,32,33,34),
        ('delete', 'page up', 'page down',
        'f1','f2','f3','f4','f5','f6','f7','f8','f9','f10','f11',
        'f12','f13','f14','f15','f16','f17','f18','f19','f20'))
] + [
    # mouse reporting (special handling done in KeyqueueTrie)
    ('[M', 'mouse'),

    # mouse reporting for SGR 1006
    ('[<', 'sgrmouse'),

    # report status response
    ('[0n', 'status ok')
]
# yapf: enable

# This is added to button value to signal mouse release by curses_display
# and raw_display when we know which button was released.  NON-STANDARD
MOUSE_RELEASE_FLAG = 2048

# This 2-bit mask is used to check if the mouse release from curses or gpm
# is a double or triple release. 00 means single click, 01 double,
# 10 triple. NON-STANDARD
MOUSE_MULTIPLE_CLICK_MASK = 1536

# This is added to button value at mouse release to differentiate between
# single, double and triple press. Double release adds this times one,
# triple release adds this times two.  NON-STANDARD
MOUSE_MULTIPLE_CLICK_FLAG = 512

# xterm adds this to the button value to signal a mouse drag event
MOUSE_DRAG_FLAG = 32


class KeyqueueTrie(object):
    def __init__(self, sequences):
        self.data = {}
        for s, result in sequences:
            assert type(result) != dict
            self.add(self.data, s, result)

    def add(self, root, s, result):
        assert type(root) == dict, "trie conflict detected"
        assert len(s) > 0, "trie conflict detected"

        if ord(s[0]) in root:
            return self.add(root[ord(s[0])], s[1:], result)
        if len(s) > 1:
            d = {}
            root[ord(s[0])] = d
            return self.add(d, s[1:], result)
        root[ord(s)] = result

    def get(self, keys, more_available):

        result = self.get_recurse(self.data, keys, more_available)
        if not result:
            result = self.read_cursor_position(keys, more_available)
        return result

    def get_recurse(self, root, keys, more_available):
        if type(root) != dict:
            if root == "mouse":
                return self.read_mouse_info(keys, more_available)
            elif root == "sgrmouse":
                return self.read_sgrmouse_info(keys, more_available)
            return (root, keys)
        if not keys:
            # get more keys
            if more_available:
                raise MoreInputRequired()
            return None
        if keys[0] not in root:
            return None
        return self.get_recurse(root[keys[0]], keys[1:], more_available)

    def read_mouse_info(self, keys, more_available):
        if len(keys) < 3:
            if more_available:
                raise MoreInputRequired()
            return None

        b = keys[0] - 32
        x, y = (keys[1] - 33) % 256, (keys[2] - 33) % 256  # supports 0-255

        prefix = ""
        if b & 4:
            prefix = prefix + "shift "
        if b & 8:
            prefix = prefix + "meta "
        if b & 16:
            prefix = prefix + "ctrl "
        if (b & MOUSE_MULTIPLE_CLICK_MASK) >> 9 == 1:
            prefix = prefix + "double "
        if (b & MOUSE_MULTIPLE_CLICK_MASK) >> 9 == 2:
            prefix = prefix + "triple "

        # 0->1, 1->2, 2->3, 64->4, 65->5
        button = ((b & 64) // 64 * 3) + (b & 3) + 1

        if b & 3 == 3:
            action = "release"
            button = 0
        elif b & MOUSE_RELEASE_FLAG:
            action = "release"
        elif b & MOUSE_DRAG_FLAG:
            action = "drag"
        elif b & MOUSE_MULTIPLE_CLICK_MASK:
            action = "click"
        else:
            action = "press"

        return ((prefix + "mouse " + action, button, x, y), keys[3:])

    def read_sgrmouse_info(self, keys, more_available):
        # Helpful links:
        # https://stackoverflow.com/questions/5966903/how-to-get-mousemove-and-mouseclick-in-bash
        # http://invisible-island.net/xterm/ctlseqs/ctlseqs.pdf

        if not keys:
            if more_available:
                raise MoreInputRequired()
            return None

        value = ""
        pos_m = 0
        found_m = False
        for k in keys:
            value = value + chr(k)
            if (k is ord("M")) or (k is ord("m")):
                found_m = True
                break
            pos_m += 1
        if not found_m:
            if more_available:
                raise MoreInputRequired()
            return None

        (b, x, y) = value[:-1].split(";")

        # shift, meta, ctrl etc. is not communicated on my machine, so I
        # can't and won't be able to add support for it.
        # Double and triple clicks are not supported as well. They can be
        # implemented by using a timer. This timer can check if the last
        # registered click is below a certain threshold. This threshold
        # is normally set in the operating system itself, so setting one
        # here will cause an inconsistent behaviour. I do not plan to use
        # that feature, so I won't implement it.

        button = ((int(b) & 64) // 64 * 3) + (int(b) & 3) + 1
        x = int(x) - 1
        y = int(y) - 1

        if value[-1] == "M":
            if int(b) & MOUSE_DRAG_FLAG:
                action = "drag"
            else:
                action = "press"
        else:
            action = "release"

        return (("mouse " + action, button, x, y), keys[pos_m + 1 :])

    def read_cursor_position(self, keys, more_available):
        """
        Interpret cursor position information being sent by the
        user's terminal.  Returned as ('cursor position', x, y)
        where (x, y) == (0, 0) is the top left of the screen.
        """
        if not keys:
            if more_available:
                raise MoreInputRequired()
            return None
        if keys[0] != ord("["):
            return None
        # read y value
        y = 0
        i = 1
        for k in keys[i:]:
            i += 1
            if k == ord(";"):
                if not y:
                    return None
                else:
                    break
            if k < ord("0") or k > ord("9"):
                return None
            if not y and k == ord("0"):
                return None
            y = y * 10 + k - ord("0")
        if not keys[i:]:
            if more_available:
                raise MoreInputRequired()
            return None
        # read x value
        x = 0
        for k in keys[i:]:
            i += 1
            if k == ord("R"):
                return None if not x else (("cursor position", x - 1, y - 1), keys[i:])
            if k < ord("0") or k > ord("9"):
                return None
            if not x and k == ord("0"):
                return None
            x = x * 10 + k - ord("0")
        if not keys[i:] and more_available:
            raise MoreInputRequired()
        return None


# ===============================================
# Build the input trie from input_sequences list
input_trie = KeyqueueTrie(input_sequences)
# ===============================================

ESC = "\x1b"
MOUSE_TRACKING_ON = ESC + "[?1000h" + ESC + "[?1002h" + ESC + "[?1006h"
MOUSE_TRACKING_OFF = ESC + "[?1006l" + ESC + "[?1002l" + ESC + "[?1000l"

_keyconv = {
    -1: None,
    8: "backspace",
    9: "tab",
    10: "enter",
    13: "enter",
    127: "backspace",
    # curses-only keycodes follow..  (XXX: are these used anymore?)
    258: "down",
    259: "up",
    260: "left",
    261: "right",
    262: "home",
    263: "backspace",
    265: "f1",
    266: "f2",
    267: "f3",
    268: "f4",
    269: "f5",
    270: "f6",
    271: "f7",
    272: "f8",
    273: "f9",
    274: "f10",
    275: "f11",
    276: "f12",
    277: "shift f1",
    278: "shift f2",
    279: "shift f3",
    280: "shift f4",
    281: "shift f5",
    282: "shift f6",
    283: "shift f7",
    284: "shift f8",
    285: "shift f9",
    286: "shift f10",
    287: "shift f11",
    288: "shift f12",
    330: "delete",
    331: "insert",
    338: "page down",
    339: "page up",
    343: "enter",  # on numpad
    350: "5",  # on numpad
    360: "end",
}


def process_keyqueue(codes, more_available):
    """
    codes -- list of key codes
    more_available -- if True then raise MoreInputRequired when in the
        middle of a character sequence (escape/utf8/wide) and caller
        will attempt to send more key codes on the next call.
    returns (list of input, list of remaining key codes).
    """
    code = codes[0]
    if code >= 32 and code <= 126:
        key = chr(code)
        return [key], codes[1:]
    if code in _keyconv:
        return [_keyconv[code]], codes[1:]
    if code > 0 and code < 27:
        return ["ctrl %s" % chr(ord("a") + code - 1)], codes[1:]
    if code > 27 and code < 32:
        return ["ctrl %s" % chr(ord("A") + code - 1)], codes[1:]

    em = get_byte_encoding()

    if em == "wide" and code < 256 and within_double_byte(chr(code), 0, 0):
        if not codes[1:]:
            if more_available:
                raise MoreInputRequired()
        if codes[1:] and codes[1] < 256:
            db = chr(code) + chr(codes[1])
            if within_double_byte(db, 0, 1):
                return [db], codes[2:]
    if em == "utf8" and code > 127 and code < 256:
        if code & 0xE0 == 0xC0:  # 2-byte form
            need_more = 1
        elif code & 0xF0 == 0xE0:  # 3-byte form
            need_more = 2
        elif code & 0xF8 == 0xF0:  # 4-byte form
            need_more = 3
        else:
            return ["<%d>" % code], codes[1:]

        for i in range(need_more):
            if len(codes) - 1 <= i:
                if more_available:
                    raise MoreInputRequired()
                else:
                    return ["<%d>" % code], codes[1:]
            k = codes[i + 1]
            if k > 256 or k & 0xC0 != 0x80:
                return ["<%d>" % code], codes[1:]

        s = bytes(codes[: need_more + 1])

        assert isinstance(s, bytes)
        try:
            return [s.decode("utf-8")], codes[need_more + 1 :]
        except UnicodeDecodeError:
            return ["<%d>" % code], codes[1:]

    if code > 127 and code < 256:
        key = chr(code)
        return [key], codes[1:]
    if code != 27:
        return ["<%d>" % code], codes[1:]

    result = input_trie.get(codes[1:], more_available)

    if result is not None:
        result, remaining_codes = result
        return [result], remaining_codes

    if codes[1:]:
        # Meta keys -- ESC+Key form
        run, remaining_codes = process_keyqueue(codes[1:], more_available)
        if is_mouse_event(run[0]):
            return ["esc"] + run, remaining_codes
        if run[0] == "esc" or run[0].find("meta ") >= 0:
            return ["esc"] + run, remaining_codes
        return ["meta " + run[0]] + run[1:], remaining_codes

    return ["esc"], codes[1:]


class InputTerminal(object):
    def __init__(self):
        super(InputTerminal, self).__init__()
        self._signal_keys_set = False
        self._old_signal_keys = None

    def tty_signal_keys(
        self, intr=None, quit=None, start=None, stop=None, susp=None, fileno=None
    ):
        """
        Read and/or set the tty's signal character settings.
        This function returns the current settings as a tuple.
        Use the string 'undefined' to unmap keys from their signals.
        The value None is used when no change is being made.
        Setting signal keys is done using the integer ascii
        code for the key, eg.  3 for CTRL+C.
        If this function is called after start() has been called
        then the original settings will be restored when stop()
        is called.
        """
        if fileno is None:
            fileno = sys.stdin.fileno()
        if not os.isatty(fileno):
            return

        tattr = termios.tcgetattr(fileno)
        sattr = tattr[6]
        skeys = (
            sattr[termios.VINTR],
            sattr[termios.VQUIT],
            sattr[termios.VSTART],
            sattr[termios.VSTOP],
            sattr[termios.VSUSP],
        )

        if intr == "undefined":
            intr = 0
        if quit == "undefined":
            quit = 0
        if start == "undefined":
            start = 0
        if stop == "undefined":
            stop = 0
        if susp == "undefined":
            susp = 0

        if intr is not None:
            tattr[6][termios.VINTR] = intr
        if quit is not None:
            tattr[6][termios.VQUIT] = quit
        if start is not None:
            tattr[6][termios.VSTART] = start
        if stop is not None:
            tattr[6][termios.VSTOP] = stop
        if susp is not None:
            tattr[6][termios.VSUSP] = susp

        if (
            intr is not None
            or quit is not None
            or start is not None
            or stop is not None
            or susp is not None
        ):
            termios.tcsetattr(fileno, termios.TCSADRAIN, tattr)
            self._signal_keys_set = True

        return skeys


class PosixInput(InputTerminal):
    def __init__(self, input=sys.stdin, output=sys.stdout) -> None:
        super().__init__()

        self._keyqueue = []
        self.prev_input_resize = 0
        self._partial_codes = None
        self._input_timeout = None
        self.set_input_timeouts()
        self._next_timeout = None
        self._resized = False

        self.gpm_mev = None
        self.gpm_event_pending = False
        self._mouse_tracking_enabled = False
        self.signal_handler_setter = signal.signal

        # Our connections to the world
        self._term_output_file = output
        self._term_input_file = input

        # pipe for signalling external event loops about resize events
        self._resize_pipe_rd, self._resize_pipe_wr = os.pipe()
        fcntl.fcntl(self._resize_pipe_rd, fcntl.F_SETFL, os.O_NONBLOCK)

    def _input_fileno(self):
        """
        Returns the fileno of the input stream, or None if it doesn't have one.
        A stream without a fileno can't participate in whatever.
        """
        if hasattr(self._term_input_file, "fileno"):
            return self._term_input_file.fileno()
        else:
            return None

    def set_input_timeouts(self, max_wait=None, complete_wait=0.125, resize_wait=0.125):
        """
        Set the get_input timeout values.  All values are in floating
        point numbers of seconds.
        max_wait -- amount of time in seconds to wait for input when
            there is no input pending, wait forever if None
        complete_wait -- amount of time in seconds to wait when
            get_input detects an incomplete escape sequence at the
            end of the available input
        resize_wait -- amount of time in seconds to wait for more input
            after receiving two screen resize requests in a row to
            stop Urwid from consuming 100% cpu during a gradual
            window resize operation
        """
        self.max_wait = max_wait
        if max_wait is not None:
            if self._next_timeout is None:
                self._next_timeout = max_wait
            else:
                self._next_timeout = min(self._next_timeout, self.max_wait)
        self.complete_wait = complete_wait
        self.resize_wait = resize_wait

    def _sigwinch_handler(self, signum, frame=None):
        """
        frame -- will always be None when the GLib event loop is being used.
        """

        if not self._resized:
            os.write(self._resize_pipe_wr, B("R"))
        self._resized = True

    def signal_init(self):
        """
        Called in the startup of run wrapper to set the SIGWINCH
        signal handlers.
        Override this function to call from main thread in threaded
        applications.
        """
        self.signal_handler_setter(signal.SIGWINCH, self._sigwinch_handler)

    def signal_restore(self):
        """
        Called in the finally block of run wrapper to restore the
        SIGWINCH signal handlers.
        Override this function to call from main thread in threaded
        applications.
        """
        self.signal_handler_setter(signal.SIGCONT, signal.SIG_DFL)

    def set_mouse_tracking(self, enable=True):
        """
        Enable (or disable) mouse tracking.
        After calling this function get_input will include mouse
        click events along with keystrokes.
        """
        enable = bool(enable)
        if enable == self._mouse_tracking_enabled:
            return

        self._mouse_tracking(enable)
        self._mouse_tracking_enabled = enable

    def _mouse_tracking(self, enable):
        if enable:
            self.write(MOUSE_TRACKING_ON)
            self._start_gpm_tracking()
        else:
            self.write(MOUSE_TRACKING_OFF)
            self._stop_gpm_tracking()

    def _start_gpm_tracking(self):
        if not os.path.isfile("/usr/bin/mev"):
            return
        if not os.environ.get("TERM", "").lower().startswith("linux"):
            return
        if not Popen:
            return
        m = Popen(
            ["/usr/bin/mev", "-e", "158"], stdin=PIPE, stdout=PIPE, close_fds=True
        )
        fcntl.fcntl(m.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)
        self.gpm_mev = m

    def _stop_gpm_tracking(self):
        if not self.gpm_mev:
            return
        os.kill(self.gpm_mev.pid, signal.SIGINT)
        os.waitpid(self.gpm_mev.pid, 0)
        self.gpm_mev = None

    def start(self):
        fd = self._input_fileno()
        if fd is not None and os.isatty(fd):
            self._old_termios_settings = termios.tcgetattr(fd)
            tty.setcbreak(fd)

        self.signal_init()
        self._next_timeout = self.max_wait

        if not self._signal_keys_set:
            self._old_signal_keys = self.tty_signal_keys(fileno=fd)

        # restore mouse tracking to previous state
        self._mouse_tracking(self._mouse_tracking_enabled)

    def stop(self):
        self.signal_restore()

        fd = self._input_fileno()
        if fd is not None and os.isatty(fd):
            termios.tcsetattr(fd, termios.TCSADRAIN, self._old_termios_settings)

        self._mouse_tracking(False)

    def _wait_for_input_ready(self, timeout):
        ready = None
        fd_list = []
        fd = self._input_fileno()
        if fd is not None:
            fd_list.append(fd)
        while True:
            try:
                if timeout is None:
                    ready, w, err = select.select(fd_list, [], fd_list)
                else:
                    ready, w, err = select.select(fd_list, [], fd_list, timeout)
                break
            except select.error as e:
                if e.args[0] != 4:
                    raise
                if self._resized:
                    ready = []
                    break
        return ready

    def _getch(self, timeout):
        ready = self._wait_for_input_ready(timeout)
        fd = self._input_fileno()
        if fd is not None and fd in ready:
            return ord(os.read(fd, 1))
        return -1

    def _getch_nodelay(self):
        return self._getch(0)

    def _get_keyboard_codes(self):
        codes = []
        while True:
            code = self._getch_nodelay()
            if code < 0:
                break
            codes.append(code)
        return codes

    def _encode_gpm_event(self):
        self.gpm_event_pending = False
        s = self.gpm_mev.stdout.readline().decode("ascii")
        l = s.split(",")
        if len(l) != 6:
            # unexpected output, stop tracking
            self._stop_gpm_tracking()
            return []
        ev, x, y, ign, b, m = s.split(",")
        ev = int(ev.split("x")[-1], 16)
        x = int(x.split(" ")[-1])
        y = int(y.lstrip().split(" ")[0])
        b = int(b.split(" ")[-1])
        m = int(m.split("x")[-1].rstrip(), 16)

        # convert to xterm-like escape sequence

        last = next = self.last_bstate
        l = []

        mod = 0
        if m & 1:
            mod |= 4  # shift
        if m & 10:
            mod |= 8  # alt
        if m & 4:
            mod |= 16  # ctrl

        def append_button(b):
            b |= mod
            l.extend([27, ord("["), ord("M"), b + 32, x + 32, y + 32])

        if ev == 20 or ev == 36 or ev == 52:  # press
            if b & 4 and last & 1 == 0:
                append_button(0)
                next |= 1
            if b & 2 and last & 2 == 0:
                append_button(1)
                next |= 2
            if b & 1 and last & 4 == 0:
                append_button(2)
                next |= 4
        elif ev == 146:  # drag
            if b & 4:
                append_button(0 + MOUSE_DRAG_FLAG)
            elif b & 2:
                append_button(1 + MOUSE_DRAG_FLAG)
            elif b & 1:
                append_button(2 + MOUSE_DRAG_FLAG)
        else:  # release
            if b & 4 and last & 1:
                append_button(0 + MOUSE_RELEASE_FLAG)
                next &= ~1
            if b & 2 and last & 2:
                append_button(1 + MOUSE_RELEASE_FLAG)
                next &= ~2
            if b & 1 and last & 4:
                append_button(2 + MOUSE_RELEASE_FLAG)
                next &= ~4
        if ev == 40:  # double click (release)
            if b & 4 and last & 1:
                append_button(0 + MOUSE_MULTIPLE_CLICK_FLAG)
            if b & 2 and last & 2:
                append_button(1 + MOUSE_MULTIPLE_CLICK_FLAG)
            if b & 1 and last & 4:
                append_button(2 + MOUSE_MULTIPLE_CLICK_FLAG)
        elif ev == 52:
            if b & 4 and last & 1:
                append_button(0 + MOUSE_MULTIPLE_CLICK_FLAG * 2)
            if b & 2 and last & 2:
                append_button(1 + MOUSE_MULTIPLE_CLICK_FLAG * 2)
            if b & 1 and last & 4:
                append_button(2 + MOUSE_MULTIPLE_CLICK_FLAG * 2)

        self.last_bstate = next
        return l

    def _get_gpm_codes(self):
        codes = []
        try:
            while self.gpm_mev is not None and self.gpm_event_pending:
                codes.extend(self._encode_gpm_event())
        except IOError as e:
            if e.args[0] != 11:
                raise
        return codes

    def get_available_raw_input(self):
        """
        Return any currently-available input.  Does not block.
        This method is only used by the default `hook_event_loop`
        implementation; you can safely ignore it if you implement your own.
        """
        codes = self._get_gpm_codes() + self._get_keyboard_codes()

        if self._partial_codes:
            codes = self._partial_codes + codes
            self._partial_codes = None

        # clean out the pipe used to signal external event loops
        # that a resize has occurred
        try:
            while True:
                os.read(self._resize_pipe_rd, 1)
        except OSError:
            pass

        return codes

    def parse_input(self, event_loop, callback, codes, wait_for_more=True):
        """
        Read any available input from get_available_raw_input, parses it into
        keys, and calls the given callback.
        The current implementation tries to avoid any assumptions about what
        the screen or event loop look like; it only deals with parsing keycodes
        and setting a timeout when an incomplete one is detected.
        `codes` should be a sequence of keycodes, i.e. bytes.  A bytearray is
        appropriate, but beware of using bytes, which only iterates as integers
        on Python 3.
        """
        # Note: event_loop may be None for 100% synchronous support, only used
        # by get_input.  Not documented because you shouldn't be doing it.
        if self._input_timeout and event_loop:
            event_loop.remove_alarm(self._input_timeout)
            self._input_timeout = None

        original_codes = codes
        processed = []
        try:
            while codes:
                run, codes = process_keyqueue(codes, wait_for_more)
                processed.extend(run)
        except MoreInputRequired:
            # Set a timer to wait for the rest of the input; if it goes off
            # without any new input having come in, use the partial input
            k = len(original_codes) - len(codes)
            processed_codes = original_codes[:k]
            self._partial_codes = codes

            def _parse_incomplete_input():
                self._input_timeout = None
                self._partial_codes = None
                self.parse_input(event_loop, callback, codes, wait_for_more=False)

            if event_loop:
                self._input_timeout = event_loop.alarm(
                    self.complete_wait, _parse_incomplete_input
                )
            else:
                # When there is no `eventloop`, and turn on `wait more`.
                run, codes = self.parse_input(
                    event_loop, callback, codes, wait_for_more=False
                )
                processed.extend(run)
                processed_codes = original_codes

        else:
            processed_codes = original_codes
            self._partial_codes = None

        if self._resized:
            processed.append("window resize")
            self._resized = False

        if callback:
            callback(processed, processed_codes)
        else:
            # For get_input
            return processed, processed_codes

    def get_input(self, raw_keys=False):
        """Return pending input as a list.
        raw_keys -- return raw keycodes as well as translated versions
        This function will immediately return all the input since the
        last time it was called.  If there is no input pending it will
        wait before returning an empty list.  The wait time may be
        configured with the set_input_timeouts function.
        If raw_keys is False (default) this function will return a list
        of keys pressed.  If raw_keys is True this function will return
        a ( keys pressed, raw keycodes ) tuple instead.
        Examples of keys returned:
        * ASCII printable characters:  " ", "a", "0", "A", "-", "/"
        * ASCII control characters:  "tab", "enter"
        * Escape sequences:  "up", "page up", "home", "insert", "f1"
        * Key combinations:  "shift f1", "meta a", "ctrl b"
        * Window events:  "window resize"
        When a narrow encoding is not enabled:
        * "Extended ASCII" characters:  "\\xa1", "\\xb2", "\\xfe"
        When a wide encoding is enabled:
        * Double-byte characters:  "\\xa1\\xea", "\\xb2\\xd4"
        When utf8 encoding is enabled:
        * Unicode characters: u"\\u00a5", u'\\u253c"
        Examples of mouse events returned:
        * Mouse button press: ('mouse press', 1, 15, 13),
                              ('meta mouse press', 2, 17, 23)
        * Mouse drag: ('mouse drag', 1, 16, 13),
                      ('mouse drag', 1, 17, 13),
                      ('ctrl mouse drag', 1, 18, 13)
        * Mouse button release: ('mouse release', 0, 18, 13),
                                ('ctrl mouse release', 0, 17, 23)
        """

        self._wait_for_input_ready(self._next_timeout)
        keys, raw = self.parse_input(None, None, self.get_available_raw_input())

        # Avoid pegging CPU at 100% when slowly resizing
        if keys == ["window resize"] and self.prev_input_resize:
            while True:
                self._wait_for_input_ready(self.resize_wait)
                keys, raw2 = self.parse_input(
                    None, None, self.get_available_raw_input()
                )
                raw += raw2
                # if not keys:
                #    keys, raw2 = self._get_input(
                #        self.resize_wait)
                #    raw += raw2
                if keys != ["window resize"]:
                    break
            if keys[-1:] != ["window resize"]:
                keys.append("window resize")

        if keys == ["window resize"]:
            self.prev_input_resize = 2
        elif self.prev_input_resize == 2 and not keys:
            self.prev_input_resize = 1
        else:
            self.prev_input_resize = 0

        if raw_keys:
            return keys, raw
        return keys

    def write(self, data):
        """Write some data to the terminal.
        You may wish to override this if you're using something other than
        regular files for input and output.
        """
        self._term_output_file.write(data)

    def flush(self):
        """Flush the output buffer.
        You may wish to override this if you're using something other than
        regular files for input and output.
        """
        self._term_output_file.flush()


if __name__ == "__main__":
    handle = PosixInput()
    # handle.set_input_timeouts(0.125)
    handle.start()
    handle.set_mouse_tracking()
    handle.set_input_timeouts(0.125)
    while True:
        res = handle.get_input(raw_keys=True)
        print(res)



from typing import Dict

# https://github.com/github/gitignore

_android: str = r"""
# Gradle files
.gradle/
build/

# Local configuration file (sdk path, etc)
local.properties

# Log/OS Files
*.log

# Android Studio generated files and folders
captures/
.externalNativeBuild/
.cxx/
*.apk
output.json

# IntelliJ
*.iml
.idea/
misc.xml
deploymentTargetDropDown.xml
render.experimental.xml

# Keystore files
*.jks
*.keystore

# Google Services (e.g. APIs or Firebase)
google-services.json

# Android Profiling
*.hprof
"""

_cpp: str = r"""
# Prerequisites
*.d

# Compiled Object files
*.slo
*.lo
*.o
*.obj

# Precompiled Headers
*.gch
*.pch

# Compiled Dynamic libraries
*.so
*.dylib
*.dll

# Fortran module files
*.mod
*.smod

# Compiled Static libraries
*.lai
*.la
*.a
*.lib

# Executables
*.exe
*.out
*.app
"""

_c: str = r"""
# Prerequisites
*.d

# Object files
*.o
*.ko
*.obj
*.elf

# Linker output
*.ilk
*.map
*.exp

# Precompiled Headers
*.gch
*.pch

# Libraries
*.lib
*.a
*.la
*.lo

# Shared objects (inc. Windows DLLs)
*.dll
*.so
*.so.*
*.dylib

# Executables
*.exe
*.out
*.app
*.i*86
*.x86_64
*.hex

# Debug files
*.dSYM/
*.su
*.idb
*.pdb

# Kernel Module Compile Results
*.mod*
*.cmd
.tmp_versions/
modules.order
Module.symvers
Mkfile.old
dkms.conf
"""

_cmake: str = r"""
CMakeLists.txt.user
CMakeCache.txt
CMakeFiles
CMakeScripts
Testing
Makefile
cmake_install.cmake
install_manifest.txt
compile_commands.json
CTestTestfile.cmake
_deps
"""

_dart: str = r"""
# See https://www.dartlang.org/guides/libraries/private-files

# Files and directories created by pub
.dart_tool/
.packages
build/
# If you're building an application, you may want to check-in your pubspec.lock
pubspec.lock

# Directory created by dartdoc
# If you don't generate documentation locally you can remove this line.
doc/api/

# dotenv environment variables file
.env*

# Avoid committing generated Javascript files:
*.dart.js
*.info.json      # Produced by the --dump-info flag.
*.js             # When generated by dart2js. Don't specify *.js if your
                 # project includes source files written in JavaScript.
*.js_
*.js.deps
*.js.map

.flutter-plugins
.flutter-plugins-dependencies
"""

_elisp: str = r"""
# Compiled
*.elc

# Packaging
.cask

# Backup files
*~

# Undo-tree save-files
*.~undo-tree
"""

_gitbook: str = r"""\
# Node rules:
## Grunt intermediate storage (http://gruntjs.com/creating-plugins#storing-task-files)
.grunt

## Dependency directory
## Commenting this out is preferred by some people, see
## https://docs.npmjs.com/misc/faq#should-i-check-my-node_modules-folder-into-git
node_modules

# Book build output
_book

# eBook build output
*.epub
*.mobi
*.pdf
"""

_go: str = r"""
# If you prefer the allow list template instead of the deny list, see community template:
# https://github.com/github/gitignore/blob/main/community/Golang/Go.AllowList.gitignore
#
# Binaries for programs and plugins
*.exe
*.exe~
*.dll
*.so
*.dylib

# Test binary, built with `go test -c`
*.test

# Output of the go coverage tool, specifically when used with LiteIDE
*.out

# Dependency directories (remove the comment below to include it)
# vendor/

# Go workspace file
go.work
"""

_java: str = r"""
# Compiled class file
*.class

# Log file
*.log

# BlueJ files
*.ctxt

# Mobile Tools for Java (J2ME)
.mtj.tmp/

# Package Files #
*.jar
*.war
*.nar
*.ear
*.zip
*.tar.gz
*.rar

# virtual machine crash logs, see http://www.java.com/en/download/help/error_hotspot.xml
hs_err_pid*
replay_pid*
"""

_lua: str = r"""
# Compiled Lua sources
luac.out

# luarocks build files
*.src.rock
*.zip
*.tar.gz

# Object files
*.o
*.os
*.ko
*.obj
*.elf

# Precompiled Headers
*.gch
*.pch

# Libraries
*.lib
*.a
*.la
*.lo
*.def
*.exp

# Shared objects (inc. Windows DLLs)
*.dll
*.so
*.so.*
*.dylib

# Executables
*.exe
*.out
*.app
*.i*86
*.x86_64
*.hex
"""

_maven: str = r"""
target/
pom.xml.tag
pom.xml.releaseBackup
pom.xml.versionsBackup
pom.xml.next
release.properties
dependency-reduced-pom.xml
buildNumber.properties
.mvn/timing.properties
# https://github.com/takari/maven-wrapper#usage-without-binary-jar
.mvn/wrapper/maven-wrapper.jar

# Eclipse m2e generated files
# Eclipse Core
.project
# JDT-specific (Eclipse Java Development Tools)
.classpath
"""

_node: str = r"""
# Logs
logs
*.log
npm-debug.log*
yarn-debug.log*
yarn-error.log*
lerna-debug.log*
.pnpm-debug.log*

# Diagnostic reports (https://nodejs.org/api/report.html)
report.[0-9]*.[0-9]*.[0-9]*.[0-9]*.json

# Runtime data
pids
*.pid
*.seed
*.pid.lock

# Directory for instrumented libs generated by jscoverage/JSCover
lib-cov

# Coverage directory used by tools like istanbul
coverage
*.lcov

# nyc test coverage
.nyc_output

# Grunt intermediate storage (https://gruntjs.com/creating-plugins#storing-task-files)
.grunt

# Bower dependency directory (https://bower.io/)
bower_components

# node-waf configuration
.lock-wscript

# Compiled binary addons (https://nodejs.org/api/addons.html)
build/Release

# Dependency directories
node_modules/
jspm_packages/

# Snowpack dependency directory (https://snowpack.dev/)
web_modules/

# TypeScript cache
*.tsbuildinfo

# Optional npm cache directory
.npm

# Optional eslint cache
.eslintcache

# Optional stylelint cache
.stylelintcache

# Microbundle cache
.rpt2_cache/
.rts2_cache_cjs/
.rts2_cache_es/
.rts2_cache_umd/

# Optional REPL history
.node_repl_history

# Output of 'npm pack'
*.tgz

# Yarn Integrity file
.yarn-integrity

# dotenv environment variable files
.env
.env.development.local
.env.test.local
.env.production.local
.env.local

# parcel-bundler cache (https://parceljs.org/)
.cache
.parcel-cache

# Next.js build output
.next
out

# Nuxt.js build / generate output
.nuxt
dist

# Gatsby files
.cache/
# Comment in the public line in if your project uses Gatsby and not Next.js
# https://nextjs.org/blog/next-9-1#public-directory-support
# public

# vuepress build output
.vuepress/dist

# vuepress v2.x temp and cache directory
.temp
.cache

# Docusaurus cache and generated files
.docusaurus

# Serverless directories
.serverless/

# FuseBox cache
.fusebox/

# DynamoDB Local files
.dynamodb/

# TernJS port file
.tern-port

# Stores VSCode versions used for testing VSCode extensions
.vscode-test

# yarn v2
.yarn/cache
.yarn/unplugged
.yarn/build-state.yml
.yarn/install-state.gz
.pnp.*
"""

_object_c: str = r"""
# Logs
logs
*.log
npm-debug.log*
yarn-debug.log*
yarn-error.log*
lerna-debug.log*
.pnpm-debug.log*

# Diagnostic reports (https://nodejs.org/api/report.html)
report.[0-9]*.[0-9]*.[0-9]*.[0-9]*.json

# Runtime data
pids
*.pid
*.seed
*.pid.lock

# Directory for instrumented libs generated by jscoverage/JSCover
lib-cov

# Coverage directory used by tools like istanbul
coverage
*.lcov

# nyc test coverage
.nyc_output

# Grunt intermediate storage (https://gruntjs.com/creating-plugins#storing-task-files)
.grunt

# Bower dependency directory (https://bower.io/)
bower_components

# node-waf configuration
.lock-wscript

# Compiled binary addons (https://nodejs.org/api/addons.html)
build/Release

# Dependency directories
node_modules/
jspm_packages/

# Snowpack dependency directory (https://snowpack.dev/)
web_modules/

# TypeScript cache
*.tsbuildinfo

# Optional npm cache directory
.npm

# Optional eslint cache
.eslintcache

# Optional stylelint cache
.stylelintcache

# Microbundle cache
.rpt2_cache/
.rts2_cache_cjs/
.rts2_cache_es/
.rts2_cache_umd/

# Optional REPL history
.node_repl_history

# Output of 'npm pack'
*.tgz

# Yarn Integrity file
.yarn-integrity

# dotenv environment variable files
.env
.env.development.local
.env.test.local
.env.production.local
.env.local

# parcel-bundler cache (https://parceljs.org/)
.cache
.parcel-cache

# Next.js build output
.next
out

# Nuxt.js build / generate output
.nuxt
dist

# Gatsby files
.cache/
# Comment in the public line in if your project uses Gatsby and not Next.js
# https://nextjs.org/blog/next-9-1#public-directory-support
# public

# vuepress build output
.vuepress/dist

# vuepress v2.x temp and cache directory
.temp
.cache

# Docusaurus cache and generated files
.docusaurus

# Serverless directories
.serverless/

# FuseBox cache
.fusebox/

# DynamoDB Local files
.dynamodb/

# TernJS port file
.tern-port

# Stores VSCode versions used for testing VSCode extensions
.vscode-test

# yarn v2
.yarn/cache
.yarn/unplugged
.yarn/build-state.yml
.yarn/install-state.gz
.pnp.*
"""

_python: str = r"""
# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# C extensions
*.so

# Distribution / packaging
.Python
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg
MANIFEST

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
*.py,cover
.hypothesis/
.pytest_cache/
cover/

# Translations
*.mo
*.pot

# Django stuff:
*.log
local_settings.py
db.sqlite3
db.sqlite3-journal

# Flask stuff:
instance/
.webassets-cache

# Scrapy stuff:
.scrapy

# Sphinx documentation
docs/_build/

# PyBuilder
.pybuilder/
target/

# Jupyter Notebook
.ipynb_checkpoints

# IPython
profile_default/
ipython_config.py

# pyenv
#   For a library or package, you might want to ignore these files since the code is
#   intended to run in multiple environments; otherwise, check them in:
# .python-version

# pipenv
#   According to pypa/pipenv#598, it is recommended to include Pipfile.lock in version control.
#   However, in case of collaboration, if having platform-specific dependencies or dependencies
#   having no cross-platform support, pipenv may install dependencies that don't work, or not
#   install all needed dependencies.
#Pipfile.lock

# poetry
#   Similar to Pipfile.lock, it is generally recommended to include poetry.lock in version control.
#   This is especially recommended for binary packages to ensure reproducibility, and is more
#   commonly ignored for libraries.
#   https://python-poetry.org/docs/basic-usage/#commit-your-poetrylock-file-to-version-control
#poetry.lock

# pdm
#   Similar to Pipfile.lock, it is generally recommended to include pdm.lock in version control.
#pdm.lock
#   pdm stores project-wide configurations in .pdm.toml, but it is recommended to not include it
#   in version control.
#   https://pdm.fming.dev/#use-with-ide
.pdm.toml

# PEP 582; used by e.g. github.com/David-OConnor/pyflow and github.com/pdm-project/pdm
__pypackages__/

# Celery stuff
celerybeat-schedule
celerybeat.pid

# SageMath parsed files
*.sage.py

# Environments
.env
.venv
env/
venv/
ENV/
env.bak/
venv.bak/

# Spyder project settings
.spyderproject
.spyproject

# Rope project settings
.ropeproject

# mkdocs documentation
/site

# mypy
.mypy_cache/
.dmypy.json
dmypy.json

# Pyre type checker
.pyre/

# pytype static type analyzer
.pytype/

# Cython debug symbols
cython_debug/

# PyCharm
#  JetBrains specific template is maintained in a separate JetBrains.gitignore that can
#  be found at https://github.com/github/gitignore/blob/main/Global/JetBrains.gitignore
#  and can be added to the global gitignore or merged into this file.  For a more nuclear
#  option (not recommended) you can uncomment the following to ignore the entire idea folder.
#.idea/
"""

_qt: str = r"""
# C++ objects and libs
*.slo
*.lo
*.o
*.a
*.la
*.lai
*.so
*.so.*
*.dll
*.dylib

# Qt-es
object_script.*.Release
object_script.*.Debug
*_plugin_import.cpp
/.qmake.cache
/.qmake.stash
*.pro.user
*.pro.user.*
*.qbs.user
*.qbs.user.*
*.moc
moc_*.cpp
moc_*.h
qrc_*.cpp
ui_*.h
*.qmlc
*.jsc
Makefile*
*build-*
*.qm
*.prl

# Qt unit tests
target_wrapper.*

# QtCreator
*.autosave

# QtCreator Qml
*.qmlproject.user
*.qmlproject.user.*

# QtCreator CMake
CMakeLists.txt.user*

# QtCreator 4.8< compilation database
compile_commands.json

# QtCreator local machine specific files for imported projects
*creator.user*

*_qmlcache.qrc
"""

_r: str = r"""
# History files
.Rhistory
.Rapp.history

# Session Data files
.RData
.RDataTmp

# User-specific files
.Ruserdata

# Example code in package build process
*-Ex.R

# Output files from R CMD build
/*.tar.gz

# Output files from R CMD check
/*.Rcheck/

# RStudio files
.Rproj.user/

# produced vignettes
vignettes/*.html
vignettes/*.pdf

# OAuth2 token, see https://github.com/hadley/httr/releases/tag/v0.3
.httr-oauth

# knitr and R markdown default cache directories
*_cache/
/cache/

# Temporary files created by R markdown
*.utf8.md
*.knit.md

# R Environment Variables
.Renviron

# pkgdown site
docs/

# translation temp files
po/*~

# RStudio Connect folder
rsconnect/
"""

_ros: str = r"""
devel/
logs/
build/
bin/
lib/
msg_gen/
srv_gen/
msg/*Action.msg
msg/*ActionFeedback.msg
msg/*ActionGoal.msg
msg/*ActionResult.msg
msg/*Feedback.msg
msg/*Goal.msg
msg/*Result.msg
msg/_*.py
build_isolated/
devel_isolated/

# Generated by dynamic reconfigure
*.cfgc
/cfg/cpp/
/cfg/*.py

# Ignore generated docs
*.dox
*.wikidoc

# eclipse stuff
.project
.cproject

# qcreator stuff
CMakeLists.txt.user

srv/_*.py
*.pcd
*.pyc
qtcreator-*
*.user

/planning/cfg
/planning/docs
/planning/src

*~

# Emacs
.#*

# Catkin custom files
CATKIN_IGNORE
"""

_ruby: str = r"""
*.gem
*.rbc
/.config
/coverage/
/InstalledFiles
/pkg/
/spec/reports/
/spec/examples.txt
/test/tmp/
/test/version_tmp/
/tmp/

# Used by dotenv library to load environment variables.
# .env

# Ignore Byebug command history file.
.byebug_history

## Specific to RubyMotion:
.dat*
.repl_history
build/
*.bridgesupport
build-iPhoneOS/
build-iPhoneSimulator/

## Specific to RubyMotion (use of CocoaPods):
#
# We recommend against adding the Pods directory to your .gitignore. However
# you should judge for yourself, the pros and cons are mentioned at:
# https://guides.cocoapods.org/using/using-cocoapods.html#should-i-check-the-pods-directory-into-source-control
#
# vendor/Pods/

## Documentation cache and generated files:
/.yardoc/
/_yardoc/
/doc/
/rdoc/

## Environment normalization:
/.bundle/
/vendor/bundle
/lib/bundler/man/

# for a library or gem, you might want to ignore these files since the code is
# intended to run in multiple environments; otherwise, check them in:
# Gemfile.lock
# .ruby-version
# .ruby-gemset

# unless supporting rvm < 1.11.0 or doing something fancy, ignore this:
.rvmrc

# Used by RuboCop. Remote config files pulled in from inherit_from directive.
# .rubocop-https?--*
"""

_rust: str = r"""
# Generated by Cargo
# will have compiled files and executables
debug/
target/

# Remove Cargo.lock from gitignore if creating an executable, leave it for libraries
# More information here https://doc.rust-lang.org/cargo/guide/cargo-toml-vs-cargo-lock.html
Cargo.lock

# These are backup files generated by rustfmt
**/*.rs.bk

# MSVC Windows builds of rustc generate these, which store debugging information
*.pdb
"""

_sass: str = r"""
.sass-cache/
*.css.map
*.sass.map
*.scss.map
"""

_scale: str = r"""
*.class
*.log

# virtual machine crash logs, see http://www.java.com/en/download/help/error_hotspot.xml
hs_err_pid*
"""

_swift: str = r"""
# Xcode
#
# gitignore contributors: remember to update Global/Xcode.gitignore, Objective-C.gitignore & Swift.gitignore

## User settings
xcuserdata/

## compatibility with Xcode 8 and earlier (ignoring not required starting Xcode 9)
*.xcscmblueprint
*.xccheckout

## compatibility with Xcode 3 and earlier (ignoring not required starting Xcode 4)
build/
DerivedData/
*.moved-aside
*.pbxuser
!default.pbxuser
*.mode1v3
!default.mode1v3
*.mode2v3
!default.mode2v3
*.perspectivev3
!default.perspectivev3

## Obj-C/Swift specific
*.hmap

## App packaging
*.ipa
*.dSYM.zip
*.dSYM

## Playgrounds
timeline.xctimeline
playground.xcworkspace

# Swift Package Manager
#
# Add this line if you want to avoid checking in source code from Swift Package Manager dependencies.
# Packages/
# Package.pins
# Package.resolved
# *.xcodeproj
#
# Xcode automatically generates this directory with a .xcworkspacedata file and xcuserdata
# hence it is not needed unless you have added a package configuration file to your project
# .swiftpm

.build/

# CocoaPods
#
# We recommend against adding the Pods directory to your .gitignore. However
# you should judge for yourself, the pros and cons are mentioned at:
# https://guides.cocoapods.org/using/using-cocoapods.html#should-i-check-the-pods-directory-into-source-control
#
# Pods/
#
# Add this line if you want to avoid checking in source code from the Xcode workspace
# *.xcworkspace

# Carthage
#
# Add this line if you want to avoid checking in source code from Carthage dependencies.
# Carthage/Checkouts

Carthage/Build/

# Accio dependency management
Dependencies/
.accio/

# fastlane
#
# It is recommended to not store the screenshots in the git repo.
# Instead, use fastlane to re-generate the screenshots whenever they are needed.
# For more information about the recommended setup visit:
# https://docs.fastlane.tools/best-practices/source-control/#source-control

fastlane/report.xml
fastlane/Preview.html
fastlane/screenshots/**/*.png
fastlane/test_output

# Code Injection
#
# After new code Injection tools there's a generated folder /iOSInjectionProject
# https://github.com/johnno1962/injectionforxcode

iOSInjectionProject/
"""

_unity: str = r"""
# This .gitignore file should be placed at the root of your Unity project directory
#
# Get latest from https://github.com/github/gitignore/blob/main/Unity.gitignore
#
/[Ll]ibrary/
/[Tt]emp/
/[Oo]bj/
/[Bb]uild/
/[Bb]uilds/
/[Ll]ogs/
/[Uu]ser[Ss]ettings/

# MemoryCaptures can get excessive in size.
# They also could contain extremely sensitive data
/[Mm]emoryCaptures/

# Recordings can get excessive in size
/[Rr]ecordings/

# Uncomment this line if you wish to ignore the asset store tools plugin
# /[Aa]ssets/AssetStoreTools*

# Autogenerated Jetbrains Rider plugin
/[Aa]ssets/Plugins/Editor/JetBrains*

# Visual Studio cache directory
.vs/

# Gradle cache directory
.gradle/

# Autogenerated VS/MD/Consulo solution and project files
ExportedObj/
.consulo/
*.csproj
*.unityproj
*.sln
*.suo
*.tmp
*.user
*.userprefs
*.pidb
*.booproj
*.svd
*.pdb
*.mdb
*.opendb
*.VC.db

# Unity3D generated meta files
*.pidb.meta
*.pdb.meta
*.mdb.meta

# Unity3D generated file on crash reports
sysinfo.txt

# Builds
*.apk
*.aab
*.unitypackage
*.app

# Crashlytics generated file
crashlytics-build.properties

# Packed Addressables
/[Aa]ssets/[Aa]ddressable[Aa]ssets[Dd]ata/*/*.bin*

# Temporary auto-generated Android Assets
/[Aa]ssets/[Ss]treamingAssets/aa.meta
/[Aa]ssets/[Ss]treamingAssets/aa/*
"""

_visual_studio: str = r"""
## Ignore Visual Studio temporary files, build results, and
## files generated by popular Visual Studio add-ons.
##
## Get latest from https://github.com/github/gitignore/blob/main/VisualStudio.gitignore

# User-specific files
*.rsuser
*.suo
*.user
*.userosscache
*.sln.docstates

# User-specific files (MonoDevelop/Xamarin Studio)
*.userprefs

# Mono auto generated files
mono_crash.*

# Build results
[Dd]ebug/
[Dd]ebugPublic/
[Rr]elease/
[Rr]eleases/
x64/
x86/
[Ww][Ii][Nn]32/
[Aa][Rr][Mm]/
[Aa][Rr][Mm]64/
bld/
[Bb]in/
[Oo]bj/
[Ll]og/
[Ll]ogs/

# Visual Studio 2015/2017 cache/options directory
.vs/
# Uncomment if you have tasks that create the project's static files in wwwroot
#wwwroot/

# Visual Studio 2017 auto generated files
Generated\ Files/

# MSTest test Results
[Tt]est[Rr]esult*/
[Bb]uild[Ll]og.*

# NUnit
*.VisualState.xml
TestResult.xml
nunit-*.xml

# Build Results of an ATL Project
[Dd]ebugPS/
[Rr]eleasePS/
dlldata.c

# Benchmark Results
BenchmarkDotNet.Artifacts/

# .NET Core
project.lock.json
project.fragment.lock.json
artifacts/

# ASP.NET Scaffolding
ScaffoldingReadMe.txt

# StyleCop
StyleCopReport.xml

# Files built by Visual Studio
*_i.c
*_p.c
*_h.h
*.ilk
*.meta
*.obj
*.iobj
*.pch
*.pdb
*.ipdb
*.pgc
*.pgd
*.rsp
*.sbr
*.tlb
*.tli
*.tlh
*.tmp
*.tmp_proj
*_wpftmp.csproj
*.log
*.tlog
*.vspscc
*.vssscc
.builds
*.pidb
*.svclog
*.scc

# Chutzpah Test files
_Chutzpah*

# Visual C++ cache files
ipch/
*.aps
*.ncb
*.opendb
*.opensdf
*.sdf
*.cachefile
*.VC.db
*.VC.VC.opendb

# Visual Studio profiler
*.psess
*.vsp
*.vspx
*.sap

# Visual Studio Trace Files
*.e2e

# TFS 2012 Local Workspace
$tf/

# Guidance Automation Toolkit
*.gpState

# ReSharper is a .NET coding add-in
_ReSharper*/
*.[Rr]e[Ss]harper
*.DotSettings.user

# TeamCity is a build add-in
_TeamCity*

# DotCover is a Code Coverage Tool
*.dotCover

# AxoCover is a Code Coverage Tool
.axoCover/*
!.axoCover/settings.json

# Coverlet is a free, cross platform Code Coverage Tool
coverage*.json
coverage*.xml
coverage*.info

# Visual Studio code coverage results
*.coverage
*.coveragexml

# NCrunch
_NCrunch_*
.*crunch*.local.xml
nCrunchTemp_*

# MightyMoose
*.mm.*
AutoTest.Net/

# Web workbench (sass)
.sass-cache/

# Installshield output folder
[Ee]xpress/

# DocProject is a documentation generator add-in
DocProject/buildhelp/
DocProject/Help/*.HxT
DocProject/Help/*.HxC
DocProject/Help/*.hhc
DocProject/Help/*.hhk
DocProject/Help/*.hhp
DocProject/Help/Html2
DocProject/Help/html

# Click-Once directory
publish/

# Publish Web Output
*.[Pp]ublish.xml
*.azurePubxml
# Note: Comment the next line if you want to checkin your web deploy settings,
# but database connection strings (with potential passwords) will be unencrypted
*.pubxml
*.publishproj

# Microsoft Azure Web App publish settings. Comment the next line if you want to
# checkin your Azure Web App publish settings, but sensitive information contained
# in these scripts will be unencrypted
PublishScripts/

# NuGet Packages
*.nupkg
# NuGet Symbol Packages
*.snupkg
# The packages folder can be ignored because of Package Restore
**/[Pp]ackages/*
# except build/, which is used as an MSBuild target.
!**/[Pp]ackages/build/
# Uncomment if necessary however generally it will be regenerated when needed
#!**/[Pp]ackages/repositories.config
# NuGet v3's project.json files produces more ignorable files
*.nuget.props
*.nuget.targets

# Microsoft Azure Build Output
csx/
*.build.csdef

# Microsoft Azure Emulator
ecf/
rcf/

# Windows Store app package directories and files
AppPackages/
BundleArtifacts/
Package.StoreAssociation.xml
_pkginfo.txt
*.appx
*.appxbundle
*.appxupload

# Visual Studio cache files
# files ending in .cache can be ignored
*.[Cc]ache
# but keep track of directories ending in .cache
!?*.[Cc]ache/

# Others
ClientBin/
~$*
*~
*.dbmdl
*.dbproj.schemaview
*.jfm
*.pfx
*.publishsettings
orleans.codegen.cs

# Including strong name files can present a security risk
# (https://github.com/github/gitignore/pull/2483#issue-259490424)
#*.snk

# Since there are multiple workflows, uncomment next line to ignore bower_components
# (https://github.com/github/gitignore/pull/1529#issuecomment-104372622)
#bower_components/

# RIA/Silverlight projects
Generated_Code/

# Backup & report files from converting an old project file
# to a newer Visual Studio version. Backup files are not needed,
# because we have git ;-)
_UpgradeReport_Files/
Backup*/
UpgradeLog*.XML
UpgradeLog*.htm
ServiceFabricBackup/
*.rptproj.bak

# SQL Server files
*.mdf
*.ldf
*.ndf

# Business Intelligence projects
*.rdl.data
*.bim.layout
*.bim_*.settings
*.rptproj.rsuser
*- [Bb]ackup.rdl
*- [Bb]ackup ([0-9]).rdl
*- [Bb]ackup ([0-9][0-9]).rdl

# Microsoft Fakes
FakesAssemblies/

# GhostDoc plugin setting file
*.GhostDoc.xml

# Node.js Tools for Visual Studio
.ntvs_analysis.dat
node_modules/

# Visual Studio 6 build log
*.plg

# Visual Studio 6 workspace options file
*.opt

# Visual Studio 6 auto-generated workspace file (contains which files were open etc.)
*.vbw

# Visual Studio 6 auto-generated project file (contains which files were open etc.)
*.vbp

# Visual Studio 6 workspace and project file (working project files containing files to include in project)
*.dsw
*.dsp

# Visual Studio 6 technical files
*.ncb
*.aps

# Visual Studio LightSwitch build output
**/*.HTMLClient/GeneratedArtifacts
**/*.DesktopClient/GeneratedArtifacts
**/*.DesktopClient/ModelManifest.xml
**/*.Server/GeneratedArtifacts
**/*.Server/ModelManifest.xml
_Pvt_Extensions

# Paket dependency manager
.paket/paket.exe
paket-files/

# FAKE - F# Make
.fake/

# CodeRush personal settings
.cr/personal

# Python Tools for Visual Studio (PTVS)
__pycache__/
*.pyc

# Cake - Uncomment if you are using it
# tools/**
# !tools/packages.config

# Tabs Studio
*.tss

# Telerik's JustMock configuration file
*.jmconfig

# BizTalk build output
*.btp.cs
*.btm.cs
*.odx.cs
*.xsd.cs

# OpenCover UI analysis results
OpenCover/

# Azure Stream Analytics local run output
ASALocalRun/

# MSBuild Binary and Structured Log
*.binlog

# NVidia Nsight GPU debugger configuration file
*.nvuser

# MFractors (Xamarin productivity tool) working folder
.mfractor/

# Local History for Visual Studio
.localhistory/

# Visual Studio History (VSHistory) files
.vshistory/

# BeatPulse healthcheck temp database
healthchecksdb

# Backup folder for Package Reference Convert tool in Visual Studio 2017
MigrationBackup/

# Ionide (cross platform F# VS Code tools) working folder
.ionide/

# Fody - auto-generated XML schema
FodyWeavers.xsd

# VS Code files for those working on multiple tools
.vscode/*
!.vscode/settings.json
!.vscode/tasks.json
!.vscode/launch.json
!.vscode/extensions.json
*.code-workspace

# Local History for Visual Studio Code
.history/

# Windows Installer files from build outputs
*.cab
*.msi
*.msix
*.msm
*.msp

# JetBrains Rider
*.sln.iml
"""


IGNORE_TEMPLATE: Dict[str, str] = {
    "android": _android,
    "c++": _cpp,
    "cpp": _cpp,
    "c": _c,
    "cmake": _cmake,
    "dart": _dart,
    "elisp": _elisp,
    "gitbook": _gitbook,
    "go": _go,
    "java": _java,
    "kotlin": _java,
    "lua": _lua,
    "maven": _maven,
    "node": _node,
    "object-c": _object_c,
    "python": _python,
    "qt": _qt,
    "r": _r,
    "ros": _ros,
    "ruby": _ruby,
    "rust": _rust,
    "scale": _scale,
    "sass": _sass,
    "swift": _swift,
    "unity": _unity,
    "visualstudio": _visual_studio,
}

__all__ = ("IGNORE_TEMPLATE",)



from typing import Optional
import pytest
import doctest

from pigit.interaction import StatusPanel, BranchPanel, FilePanel
import shutil


def test():
    print()

    size = shutil.get_terminal_size()
    panel = StatusPanel(widget=FilePanel())
    panel._render(size)

    branch_panel = BranchPanel()



# -*- coding:utf-8 -*-

from typing import Dict, Optional

from .bash import BashCompletion
from .zsh import ZshCompletion
from .fish import FishCompletion


supported_shell = {
    "bash": BashCompletion,
    "zsh": ZshCompletion,
    "fish": FishCompletion,
}


def shell_complete(
    complete_vars: Dict,
    script_dir: Optional[str] = None,
    shell: Optional[str] = None,
    prog: Optional[str] = None,
    script_name: Optional[str] = None,
    inject_path: Optional[str] = None,
    inject: bool = True,
) -> None:
    """Generate completion script source and try to injecting.

    Args:
        complete_vars (Dict): a dict of ~Parser serialization.
        script_dir (Optional[str], optional): where the script saved. Defaults to None.
        shell (Optional[str], optional): shell name. Defaults to None.
        prog (Optional[str], optional): cmd prog. Defaults to None.
        script_name (Optional[str], optional): completion script name. Defaults to None.
        inject_path (Optional[str], optional): where the script injecting. Defaults to None.
        inject (bool, optional): whether inject script. Defaults to True.
    """

    # check shell validable.
    shell = get_shell() if shell is None else shell.lower().strip()
    if not shell:
        print("No shell be found!")
        return

    if shell not in supported_shell:
        print(
            "shell name '{0}' is not supported, see {1}".format(
                shell, supported_shell.keys()
            )
        )
        return

    print("\n===Try to add completion ...")
    print(f":: Completion shell: {repr(shell)}")

    complete_handle = supported_shell[shell](
        prog, complete_vars, script_dir, script_name, inject_path
    )

    # try create completion file.
    completion_src = complete_handle.generate_resource()
    if not inject or not script_dir:
        print(completion_src)
        return

    if not complete_handle.write_completion(completion_src):
        print(":: Write completion script failed!")
        return None
    else:
        print(":: Write completion script success.")

    # try inject to shell config.
    try:
        injected = complete_handle.inject_into_shell()
        if injected:
            print(":: Source your shell configuration.")
        else:
            print(":: Command already exist.")
    except Exception as e:
        print(e)


def get_shell() -> str:
    """Gets the currently used shell.

    Returns:
            (str): Current shell string.
    """
    import os

    try:
        shell_string = os.environ["SHELL"]
        return shell_string.split("/")[-1].strip()
    except KeyError:
        return ""



# -*- coding:utf-8 -*-
import pytest
import doctest
from unittest.mock import patch
from pprint import pprint

from pigit.common.utils import (
    traceback_info,
    confirm,
    exec_cmd,
    async_run_cmd,
    exec_async_tasks,
)
from pigit.common.func import dynamic_default_attrs


def test_doctest():
    import pigit.common.utils

    doctest.testmod(pigit.common.utils, verbose=True)


def test_traceback_info():
    try:
        _ = int("abcd")
    except Exception:
        print(traceback_info("here is extra msg."))

    # when no traceback
    assert traceback_info() == ""


@pytest.mark.parametrize(
    ("input_value", "return_bool"),
    [
        ["", True],
        ["y", True],
        ["yes", True],
        ["n", False],
        ["no", False],
    ],
)
@patch("builtins.input")
def test_confirm(mock_input, input_value: str, return_bool: bool):
    mock_input.return_value = input_value
    assert confirm("confirm:") == return_bool


def test_exec_cmd():
    # has reply
    print(exec_cmd("pwd"))
    print(exec_cmd("which", "python3"))

    # output to shell
    print(exec_cmd("pwd", reply=False))
    print(exec_cmd("which", "python3", reply=False))

    # don't decoding
    print(exec_cmd("which", "python3", decoding=False))

    # execute error
    print(exec_cmd("xxxxxxxxx"))
    print(exec_cmd("xxxxxxxxx", cwd="xxxxxxxxxxxx"))


def test_async_cmd_func():
    print()

    code = """\
# -*- coding:utf-8 -*-

if __name__ == '__main__':
    import time

    print({0})
    time.sleep(int({0}))
    print({0})
"""
    tasks = [
        async_run_cmd(*["python3", "-c", code.format(i)], msg=f"msg {i}.")
        for i in range(5)
    ]
    pprint(tasks)
    result = exec_async_tasks(tasks)
    print(result)


class TestFunc:
    def test_dynamic_default_attrs(self):
        da = {"c": 3, "d": 4}
        f = dynamic_default_attrs(lambda a, b, c, d: (a, b, c, d), **da)

        assert f(1, 2) == (1, 2, 3, 4)
        assert f(1, 2, d=0) == (1, 2, 3, 0)
        # assert f(1, 2, 0) == (1, 2, 0, 4)

        def bp(a, b, c: int, d: int = 10):
            return (a, b, c, d)

        assert dynamic_default_attrs(bp, **da)(1, 2, 0) == (1, 2, 0, 4)



# -*- coding:utf-8 -*-

from typing import Dict, Optional, Tuple
import os, re


class ShellCompletionError(Exception):
    """ShellCompletion error class."""

    pass


class ShellCompletion(object):
    """Implement and inject help classes for completion scripts."""

    _SHELL: str  # shell name.
    _INJECT_PATH: str  # script inject path.
    _template_source: str  # script tempelate string.

    def __init__(
        self,
        prog_name: Optional[str] = None,
        complete_vars: Dict = None,
        script_dir: Optional[str] = None,
        script_name: Optional[str] = None,
        inject_path: Optional[str] = None,
    ) -> None:
        """Initialization.

        Args:
            prop (str): completion trigger command.
            complete_var (dict): complete arguments dict.
                >>> complete_vars = {
                ...     '-h': 'Display help messages',
                ...     '-v': 'Show version and exit',
                ... }
            script_dir (str): where is the completion file save.
            script_name (str, optional): completion file name. Defaults to None.
            inject_path (str, optional): script inject path.

        Raises:
            TypeError: when `complete_var` is not dict.
        """
        super(ShellCompletion, self).__init__()

        if not isinstance(complete_vars, dict):
            raise TypeError("complete_var muse be dict.")
        self.complete_vars = complete_vars

        if prog_name is not None:
            self.prog_name = prog_name
        elif inner_prog_name := complete_vars.get("prog"):
            self.prog_name = inner_prog_name
        else:
            raise ShellCompletionError("Can't get prog name anywhere.") from None

        self.script_dir = script_dir or "."
        self.script_name = script_name or "{0}_{1}_comp".format(
            self.prog_name, self._SHELL
        )
        self.inject_path = inject_path or self._INJECT_PATH

    @property
    def func_name(self) -> str:
        """The name of the shell function defined by the completion
        script.
        """

        safe_name = re.sub(r"\W*", "", self.prog_name.replace("-", "_"), re.ASCII)
        return f"_{safe_name}_completion"

    def _parse(self, args: Dict) -> Tuple:
        _arguments = []
        _positions = []
        _sub_opts = {}

        for name, prop in args.items():
            prop_type = prop.get("type")

            if prop_type == "groups":
                a, p, s = self._parse(prop["args"])
                _arguments.extend(a)
                _positions.extend(p)
            elif name == "set_defaults":
                # Special need be ignore.
                pass
            elif "args" in prop:
                a, p, s = self._parse(prop["args"])
                _sub_opts[name] = {
                    "_arguments": a,
                    "_positions": p,
                    "_sub_opts": s,
                    "help": prop.get("help", "_").replace("\n", ""),
                }
            elif name.startswith("-"):
                _arguments.append((name, prop.get("help", "_").replace("\n", "")))
            else:
                _positions.append((name, prop.get("help", "_").replace("\n", "")))

        return _arguments, _positions, _sub_opts

    def generate(self):
        """Generate script content.

        Process self.complete_var to generate completion script content part.
        Due to different shells, the generated formats are different, which
        are finally implemented by subclasses.
        """

        raise NotImplementedError()

    def generate_resource(self) -> str:
        """Generate completion scirpt.

        Generate the completion script of the corresponding shell according to
        the template.

        Returns:
            (str): completion source.
        """

        complete_content = self.generate()
        return self._template_source % {
            "func_name": self.func_name,
            "prop": self.prog_name,
            "complete_vars": complete_content,
        }

    def write_completion(self, complete_src: str) -> bool:
        """Save completion to config path.

        Args:
            complete_src (str): completion source.

        Returns:
            (str): completion full path.
        """

        if not os.path.isdir(self.script_dir):
            os.makedirs(self.script_dir, exist_ok=True)

        full_path = os.path.join(self.script_dir, self.script_name)

        try:
            with open(full_path, "w" if os.path.isfile(full_path) else "x") as f:
                for line in complete_src:
                    f.write(line)
        except Exception as e:
            return False
        else:
            return True

    def inject_into_shell(self) -> bool:
        """Try using completion script.

        Inject the load of completion script into the configuration of shell.
        If it exists in the configuration, the injection will not be repeated.
        """

        full_script_path = os.path.join(self.script_dir, self.script_name)

        # Check whether already exist.
        try:
            with open(self.inject_path, "r") as f:
                shell_conf = f.read()
        except Exception as e:
            raise ShellCompletionError(
                "Read shell config error: {0}".format(e)
            ) from None
        else:
            _re = re.compile(f"{full_script_path}[^\\s]*")
            files = _re.findall(shell_conf)

        injected = bool(files)
        if injected:
            return False

        try:
            # Inject.
            with open(self.inject_path, "a") as f:
                f.write(f"source {full_script_path}")
        except Exception as e:
            raise ShellCompletionError(f"Inject error: {str(e)}") from None
        return True



from typing import TYPE_CHECKING, Callable, List, Optional, IO
import os, cmd
import functools

from plenty import get_console

if TYPE_CHECKING:
    from pigit.gitlib.processor import ShortGitter


class PigitShell(cmd.Cmd):
    intro: str = get_console().render_str(
        "b`Welcome come PIGIT shell.`<khaki>\n"
        "`You can use short commands directly. Input '?' to get help.`<khaki>\n"
    )
    prompt: str = "(pigit)> "

    def __init__(
        self,
        short_gitter: "ShortGitter",
        *,
        completekey: str = "tab",
        stdin: Optional[IO[str]] = None,
        stdout: Optional[IO[str]] = None,
    ) -> None:
        super().__init__(completekey, stdin, stdout)

        self.console = get_console()
        self.short_gitter = short_gitter

        for key, values in short_gitter.cmds.items():
            func_name = f"do_{key}"

            self.set_instance_method(self.make_fun(key, values.get("help")), func_name)

    # =================
    # cmd tools method
    # =================
    def make_fun(self, key: str, doc: str):
        _key = key

        def func(args: str):
            self.short_gitter.process_command(_key, args.split())

        func.__doc__ = doc

        return func

    @classmethod
    def set_instance_method(cls, func: Callable, func_name: Optional[str] = None):
        @functools.wraps(func)
        def dummy(self, *args, **kwargs):
            func(*args, **kwargs)

        if func_name is None:
            func_name = func.__name__

        setattr(cls, func_name, dummy)

    # ==================
    # print help method
    # ==================
    def default(self, line):
        """Called on an input line when the command prefix is not recognized.

        If this method is not overridden, it prints an error message and
        returns.

        """
        self.stdout.write(
            get_console().render_str(
                "`pigit shell: Invalid command '{0}', please select from`<error> "
                "`[shell, quit] or git short command.`<error>\n".format(line.split()[0])
            )
        )

    def do_help(self, arg: str):
        """List available commands with "help" or detailed help with "help cmd"."""
        if not arg:
            return super().do_help(arg)

        args: List[str] = arg.split()
        for a in args:
            self.stdout.write(f"{a}: ")
            super().do_help(a)

    # ================
    # support command
    # ================
    def do_shell(self, args: str):
        """Run a shell command.

        \rThis command is help you to run a normal terminal command in pigit shell.
        \rFor example, you can use `sh ls`<ok> to check the files of current dir.
        """
        os.system(args)

    do_sh = do_shell

    def do_all(self, args: str):
        """Show all short git cmds help."""
        self.short_gitter.print_help()



# -*- coding:utf-8 -*-

from typing import Dict, List, Optional, Tuple, Union, Generator
from collections import Counter
from pathlib import Path
import os, re, textwrap, json

from plenty.str_utils import shorten, byte_str2str
from plenty.console import Console

from pigit.common.utils import (
    adjudgment_type,
    async_run_cmd,
    exec_async_tasks,
    exec_cmd,
    get_file_icon,
)
from pigit.gitlib.model import File, Commit, Branch


class GitOptionError(Exception):
    """Error class of ~GitOption."""


class GitOption:
    """Git option class."""

    def __init__(
        self, op_path: Optional[str] = None, repo_json_path: Optional[str] = None
    ) -> None:
        self.op_path = op_path
        self.repo_json_path = (
            Path("./repos.json") if repo_json_path is None else Path(repo_json_path)
        )

        # create repo path dir.
        self.repo_json_path.parent.mkdir(parents=True, exist_ok=True)

    def update_setting(
        self, *, op_path: Optional[str] = None, repo_info_path: Optional[str] = None
    ) -> "GitOption":
        if op_path is not None:
            self.op_path = op_path
        if repo_info_path is not None:
            self.repo_json_path = repo_info_path

        return self

    # ==================
    # Basic info option
    # ==================
    @property
    def git_version(self) -> str:
        """Get Git version."""

        _, _version = exec_cmd("git --version")
        return _version.strip() or ""

    def get_repo_info(
        self, given_path: Optional[str] = None, exclude_submodule: bool = False
    ) -> Tuple[str, str]:
        """
        Get the current git repository path. If not, the path is empty.
        Get the local git config path. If not, the path is empty.

        Return:
                (tuple[str, str]): repository path, git config path.
        """

        path = given_path or self.op_path or "."
        repo_path: str = ""
        git_conf_path: str = ""

        path = os.path.abspath(path)
        if not os.path.isdir(path):
            return repo_path, git_conf_path

        err, repo_path = exec_cmd("git rev-parse --git-dir", cwd=path)
        if err:
            return repo_path, git_conf_path

        # remove useless space.
        repo_path = repo_path.strip()

        if ".git/modules/" in repo_path and not exclude_submodule:
            # this repo is submodule.
            git_conf_path = repo_path
            repo_path = repo_path.replace(".git/modules/", "")
        elif repo_path == ".git":
            repo_path = path
            git_conf_path = os.path.join(repo_path, ".git")
        else:
            git_conf_path = repo_path
            repo_path = repo_path[:-5]

        return repo_path, git_conf_path

    def get_head(self, path: Optional[str] = None) -> Optional[str]:
        """Get current repo head. Return a branch name or a commit sha string."""

        path = path or self.op_path
        _, res = exec_cmd(
            "git symbolic-ref -q --short HEAD || git describe --tags --exact-match",
            cwd=path,
        )
        if res is not None:
            res = res.rstrip()
        return res

    def get_branches(
        self,
        path: Optional[str] = None,
        include_remote: bool = False,
        plain: bool = True,
    ) -> List[str]:
        """Get repo all branch."""

        path = path or self.op_path
        include_remote = "--all" if include_remote else ""
        color = "never" if plain else "always"

        err, res = exec_cmd(f"git branch {include_remote} --color={color}", cwd=path)

        if res is None:
            return []
        return [branch[2:] for branch in res.rstrip().split("\n")]

    def get_first_pushed_commit(
        self, path: Optional[str] = None, branch_name: Optional[str] = None
    ) -> str:
        path = path or self.op_path

        if branch_name is None:
            if branches := self.get_branches(path):
                branch_name = branches[0]
            else:
                return ""

        command = "git merge-base %s %s@{u}" % (branch_name, branch_name)
        _, resp = exec_cmd(command, cwd=path)
        return resp.strip()

    def get_remotes(self, path: Optional[str] = None) -> List[str]:
        """Get repo remote url."""

        # Get remote name, exit when error.
        path = path or self.op_path
        err, res = exec_cmd("git remote show", cwd=path)

        if err:
            return []

        return res.strip().split("\n")

    def get_remote_url(
        self, path: Optional[str] = None, remote_name: Optional[str] = None
    ) -> str:
        """Get repo remote url."""

        path = path or self.op_path
        if remote_name is None:
            if remotes := self.get_remotes(path):
                remote_name = remotes[0]
            else:
                return ""

        # Get remote url, exit when error.
        err, remote_url = exec_cmd(f"git ls-remote --get-url {remote_name}", cwd=path)

        if err:
            return ""

        remote_url = remote_url[:-5]
        return remote_url

    def get_summary(self, path: Optional[str] = None, plain: bool = True) -> str:
        path = path or self.op_path
        color = "never" if plain else "always"
        err, res = exec_cmd(
            f"git shortlog --summary --numbered --color={color}", cwd=path
        )
        return res

    def get_repo_desc(
        self,
        include_part: Optional[List] = None,
        path: Optional[str] = None,
        color: bool = True,
    ) -> str:
        """Return a string of repo various information.

        Args:
                include_part (Optional[list], optional): should return info part: [path,remote,branch,log,summary]. Defaults to None.
                path (Optional[str], optional): custom repo path. Defaults to None.
                color (bool, optional): whether return with color. Defaults to True.

        Returns:
                str: desc info.
        """

        path = path or self.op_path
        error_str = "`Error getting.`<error>"
        gen = ["[b`Repository Information`]" if color else "[Repository Information]"]
        repo_path, _ = self.get_repo_info(path)

        # Get content.
        if not include_part or "path" in include_part:
            gen.append(
                f"Repository: \n\t`{repo_path}`<sky_blue>\n"
                if color
                else f"Repository: \n\t{repo_path}\n"
            )

        # Get remote url.
        if not include_part or "remote" in include_part:
            try:
                with open(f"{repo_path}/.git/config", "r") as cf:
                    config = cf.read()
            except Exception:
                remote = error_str
            else:
                res = re.findall(r"url\s=\s(.*)", config)
                remote = "\n".join(
                    [f"\ti`{x}`<sky_blue>" if color else f"\t{x}" for x in res]
                )
            gen.append("Remote: \n%s\n" % remote)

        # Get all branches.
        if not include_part or "branch" in include_part:
            branches = self.get_branches(path, include_remote=True, plain=not color)
            branches_str = (
                "\n".join(f"\t{branch}" for branch in branches)
                if branches
                else f"\t{error_str}"
            )
            gen.append("Branches: \n%s\n" % branches_str)

        # Get the lastest log.
        if not include_part or "log" in include_part:
            err, res = exec_cmd(
                f"git log --stat --oneline --decorate -1 --color={'always' if color else 'never'}",
                cwd=path,
            )
            git_log = "\t" + error_str if err else textwrap.indent(res, "\t")
            gen.append("Lastest log:\n%s\n" % git_log)

        # FIXME: will broken in a init repo.
        # Get git summary.
        # if not include_part or "summary" in include_part:
        #     summary = self.get_summary(path, not color)
        #     gen.append("Summary:\n%s\n" % summary or error_str)

        return "\n".join(gen)

    # =============
    # Special info
    # =============
    def load_branches(self, path: Optional[str] = None) -> List[Branch]:
        path = path or self.op_path
        command = (
            "git branch --sort=-committerdate "
            '--format="%(HEAD)|%(refname:short)|%(upstream:short)|%(upstream:track)" '
        )
        branches = []

        err, resp = exec_cmd(command, cwd=path)
        resp = resp.strip()
        if not resp:
            return branches

        lines = resp.split("\n")

        for line in lines:
            items = line.split("|")
            branch = Branch(
                name=items[1], pushables="?", pullables="?", is_head=items[0] == "*"
            )

            upstream_name = items[2]

            if not upstream_name:
                branches.append(branch)
                continue

            branch.upstream_name = upstream_name

            track = items[3]
            _re = re.compile(r"ahead (\d+)")
            branch.pushables = str(match[1]) if (match := _re.search(track)) else "0"
            _re = re.compile(r"behind (\d+)")
            branch.pullables = str(match[1]) if (match := _re.search(track)) else "0"
            branches.append(branch)

        return branches

    def load_log(
        self,
        branch_name: str = "",
        limit: Optional[int] = None,
        filter_path: str = "",
        arg_str: str = '--oneline --pretty=format:"%H|%at|%aN|%d|%p|%s" --abbrev=20 --date=unix',
        path: Optional[str] = None,
    ) -> str:
        path = path or self.op_path
        limit_flag = f"-{limit}" if limit else ""
        filter_flag = f"--follow -- {filter_path}" if filter_path else ""

        command = f"git log {branch_name} {arg_str} {limit_flag}  {filter_flag}"
        err, resp = exec_cmd(command, cwd=path)

        if resp is None:
            return ""
        return resp.strip()

    def load_status(
        self,
        max_width: int = 80,
        ident: int = 2,
        plain: bool = False,
        path: Optional[str] = None,
        icon: bool = False,
    ) -> List[File]:
        """Get the file tree status of GIT for processing and encapsulation.
        Args:
                max_width (int): The max length of display string.
                ident (int, option): Number of reserved blank characters in the header.
        Returns:
                (List[File]): Processed file status list.
        """

        path = path or self.op_path
        file_items = []

        err, files = exec_cmd("git status -s -u --porcelain", cwd=path)
        if err:
            return file_items
        for file in files.rstrip().split("\n"):
            if not file.strip():
                # skip blank line.
                continue

            change = file[:2]
            staged_change = file[:1]
            unstaged_change = file[1:2]
            name = file[3:]
            if name.endswith('"'):
                # may is chinese char code.
                name = byte_str2str(name[1:-1])
            untracked = change in ["??", "A ", "AM"]
            has_no_staged_change = staged_change in [" ", "U", "?"]
            has_merged_conflicts = change in ["DD", "AA", "UU", "AU", "UA", "UD", "DU"]
            has_inline_merged_conflicts = change in ["UU", "AA"]

            display_name = shorten(name, max_width - 3 - ident)

            icon_str = get_file_icon(adjudgment_type(display_name)) if icon else ""

            # color full command.
            display_str = Console.render_str(
                f"`{staged_change}`<{'bad' if has_no_staged_change else'right'}>`{unstaged_change}`<{'bad' if unstaged_change!=' ' else'right'}> {icon_str}{display_name}"
            )

            file_ = File(
                name=name,
                display_str=file if plain else display_str,
                short_status=change,
                has_staged_change=not has_no_staged_change,
                has_unstaged_change=unstaged_change != " ",
                tracked=not untracked,
                deleted=unstaged_change == "D" or staged_change == "D",
                added=unstaged_change == "A" or untracked,
                has_merged_conflicts=has_merged_conflicts,
                has_inline_merged_conflicts=has_inline_merged_conflicts,
            )

            file_items.append(file_)

        return file_items

    def load_file_diff(
        self,
        file: str,
        tracked: bool = True,
        cached: bool = False,
        plain: bool = False,
        path: str = None,
    ) -> str:
        """Gets the modification of the file.
        Args:
                file (str): file path relative to git.
                tracked (bool, optional): Defaults to True.
                cached (bool, optional): Defaults to False.
                plain (bool, optional): Whether need color. Defaults to False.
        Returns:
                (str): change string.
        """

        path = path or self.op_path
        command = "git diff --submodule --no-ext-diff {plain} {cached} {tracked} {file}"

        _plain = "--color=never" if plain else "--color=always"
        _cached = "--cached" if cached else ""
        _tracked = "--" if tracked else "--no-index -- /dev/null"

        if "->" in file:  # rename status.
            file = file.split("->")[-1].strip()

        err, res = exec_cmd(
            command.format(plain=_plain, cached=_cached, tracked=_tracked, file=file),
            cwd=path,
        )
        if err:
            return "Can't get diff."
        return res.rstrip()

    def load_commits(
        self,
        branch_name: str,
        limit: bool = True,
        filter_path: str = "",
        path: Optional[str] = None,
    ) -> List[Commit]:
        """Get the all commit of a given branch.
        Args:
                branch_name (str): want branch name.
                limit (bool): Whether to get only the latest 300.
                filter_path (str): filter dir path, default is empty.
        """

        path = path or self.op_path
        command = "git merge-base %s %s@{u}" % (branch_name, branch_name)

        _, resp = exec_cmd(command, cwd=path)
        first_pushed_commit = resp.strip()

        passed_first_pushed_commit = not first_pushed_commit
        commits: List[Commit] = []

        # Generate git command.
        limit_flag = "-300" if limit else ""
        filter_flag = f"--follow -- {filter_path}" if filter_path else ""
        command = f'git log {branch_name} --oneline --pretty=format:"%H|%at|%aN|%d|%p|%s" {limit_flag} --abbrev=20 --date=unix {filter_flag}'

        err, resp = exec_cmd(command, cwd=path)
        if err:
            return commits  # current is empty list.

        # Process data.
        for line in resp.split("\n"):
            split_ = line.split("|")

            sha = split_[0]
            unix_timestamp = int(split_[1])
            author = split_[2]
            extra_info = (split_[3]).strip()
            # parent_hashes = split_[4]
            message = "|".join(split_[5:])

            tag = []
            if extra_info:
                _re = re.compile(r"tag: ([^,\\]+)")
                if match := _re.search(extra_info):
                    tag.append(match[1])

            if sha == first_pushed_commit:
                passed_first_pushed_commit = True
            status = {True: "unpushed", False: "pushed"}[not passed_first_pushed_commit]

            commit_ = Commit(
                sha=sha,
                msg=message,
                author=author,
                unix_timestamp=unix_timestamp,
                status=status,
                extra_info=extra_info,
                tag=tag,
            )
            commits.append(commit_)

        return commits

    def load_commit_info(
        self,
        commit_sha: str,
        file_name: str = "",
        plain: bool = False,
        path: Optional[str] = None,
    ) -> str:
        """Gets the change of a file or all in a given commit.
        Args:
                commit_sha: commit id.
                file_name: file name(include full path).
                plain: whether has color.
        """

        path = path or self.op_path
        color_str = "never" if plain else "always"

        command = f"git show --color={color_str} {commit_sha} {file_name}"
        _, resp = exec_cmd(command, cwd=path)
        return resp.rstrip()

    # ===============
    # Options of git
    # ===============
    def _get_file_str(self, file) -> str:
        if isinstance(file, File):
            file_name = file.name
        elif isinstance(file, str):
            file_name = file
        else:
            raise GitOptionError("`file` only allow 'str' or 'File'.") from None

        if "->" in file_name:
            file_name = file_name.split("->")[-1].strip()

        return file_name

    def switch_file_status(self, file: Union[File, str], path: Optional[str] = None):
        path = path or self.op_path
        file_name = self._get_file_str(file)

        if file.has_merged_conflicts or file.has_inline_merged_conflicts:
            pass
        elif file.has_unstaged_change:
            exec_cmd(f"git add -- '{file_name}'", cwd=path)
        elif file.has_staged_change:
            if file.tracked:
                exec_cmd(f"git reset HEAD -- '{file_name}'", cwd=path)
            else:
                exec_cmd(f"git rm --cached --force -- '{file_name}'", cwd=path)

    def discard_file(
        self,
        file: Union[File, str],
        path: Optional[str] = None,
        tracked: Optional[bool] = None,
    ):
        path = path or self.op_path
        file_name = self._get_file_str(file)

        if tracked is None:
            if isinstance(file, File):
                tracked = file.tracked
            else:
                raise GitOptionError("Please set `tracked` or give a 'File'.") from None

        if tracked:
            exec_cmd(f"git checkout -- {file_name}", cwd=path)
        else:
            repo_path, _ = self.get_repo_info(path)
            os.remove(os.path.join(repo_path, file_name))

    def ignore_file(self, file: Union[File, str], path: Optional[str] = None):
        """Append file to `.gitignore` file."""

        path = path or self.op_path
        repo_path, _ = self.get_repo_info(path)
        file_name = self._get_file_str(file)

        with open(f"{repo_path}/.gitignore", "a+") as f:
            f.write(f"\n{file_name}")

    def checkout_branch(self, branch_name: str, path: Optional[str] = None) -> str:
        path = path or self.op_path
        err, _ = exec_cmd(f"git checkout {branch_name}", cwd=path)
        return err or "ok"

    def open_repo_in_browser(
        self,
        path: Optional[str] = None,
        branch: str = "",
        issue: str = "",
        commit: str = "",
        print: bool = False,
    ) -> bool:
        path = path or self.op_path
        remote_url = self.get_remote_url(path=path)

        if branch:
            branch = f"/tree/{branch}"
            remote_url += branch
        elif issue:
            issue = f"/issues/{issue}"
            remote_url += issue
        elif commit:
            commit = f"/commit/{commit}"
            remote_url += commit

        if print:
            return True, f"Remote URL: `{remote_url}`<sky_blue>"

        try:
            import webbrowser

            webbrowser.open(remote_url)
        except Exception as e:
            return False, f"Failed to open the repo; {e}"
        else:
            return True, "Successfully opened repo."

    # ====================
    # custom repo options
    # ====================
    @staticmethod
    def _make_repo_name(path: str, repos: List[str], name_counts: Counter) -> str:
        """
        Given a new repo `path`, create a repo name. By default, basename is used.
        If name collision exists, further include parent path name.

        Args:
            path (str): repo path.
            repos (list[str]): path list of exist repos.
            name_counts (Counter): generated default name.

        Returns:
            (str): final repo name.
        """

        name: str = os.path.basename(os.path.normpath(path))
        if name in repos or name_counts[name] > 1:
            # path has no trailing /
            par_name = os.path.basename(os.path.dirname(path))
            return os.path.join(par_name, name)
        return name

    def load_repos(self) -> Dict:
        if not self.repo_json_path.is_file():
            return {}

        with self.repo_json_path.open(mode="r") as fp:
            return json.load(fp)

    def save_repos(self, repos: Dict) -> bool:
        try:
            with self.repo_json_path.open(mode="w+") as fp:
                json.dump(repos, fp, indent=2)
                return True
        except Exception:
            return False

    def clear_repos(self) -> None:
        self.repo_json_path.unlink(missing_ok=True)

    def add_repos(self, paths: List[str], dry_run: bool = False) -> List:
        """Traverse the incoming paths. If it is not saved and is a git
        directory, add it to repos.

        Args:
            paths (list[str]): incoming paths.
            dry_run (bool, optional): Show but not really execute. Defaults to False.
            silent (bool, optional): No output. Defaults to False.
        """

        exist_repos = self.load_repos()
        exist_paths = [r["path"] for r in exist_repos.values()]

        new_git_paths = []
        for path in paths:
            repo_path, _ = self.get_repo_info(path)
            if repo_path and repo_path not in exist_paths:
                new_git_paths.append(repo_path)

        if new_git_paths and not dry_run:
            name_counts = Counter(
                os.path.basename(os.path.normpath(p)) for p in new_git_paths
            )
            new_repos = {
                self._make_repo_name(path, exist_repos, name_counts): {"path": path}
                for path in new_git_paths
            }

            self.save_repos({**exist_repos, **new_repos})

        return new_git_paths

    def rm_repos(self, repos: List[str], use_path: bool = False) -> List[Tuple]:
        exist_repos = self.load_repos()

        del_repos = []
        del_paths = []
        if use_path:
            del_repos.extend(
                repo for repo, info in exist_repos.items() if info["path"] in repos
            )
        else:
            del_repos.extend(repo for repo in repos if exist_repos.get(repo))

        for repo in del_repos:
            del_paths.append(exist_repos[repo]["path"])
            del exist_repos[repo]

        self.save_repos(exist_repos)
        return zip(del_repos, del_paths)

    def rename_repo(self, repo: str, name: str) -> Tuple[bool, str]:
        exist_repos = self.load_repos()

        if name in exist_repos:
            return False, f"'{name}' is already in use!"
        elif repo not in exist_repos:
            return False, f"'{repo}' is not a valid repo name!"
        else:
            prop = exist_repos[repo]
            del exist_repos[repo]
            exist_repos[name] = prop

            self.save_repos(exist_repos)
            return True, f"rename successful, `{repo}`->`{name}`."

    def ll_repos(self, revese: bool = False) -> Generator[List[Tuple], None, None]:
        exist_repos = self.load_repos()

        for repo_name, prop in exist_repos.items():
            repo_path = prop["path"]
            head = self.get_head(repo_path)

            # jump invalid repo.
            if head is None:
                if revese:
                    yield [
                        (repo_name, ""),
                        ("Local Path", repo_path),
                    ]

            elif not revese:
                _, unstaged = exec_cmd("git diff --stat", cwd=repo_path)
                _, staged = exec_cmd("git diff --stat --cached", cwd=repo_path)
                _, untracked = exec_cmd(
                    "git ls-files -zo --exclude-standard", cwd=repo_path
                )
                commit_hash = self.get_first_pushed_commit(
                    path=repo_path, branch_name=head
                )
                commit = self.load_log(
                    limit=1,
                    arg_str="--format='%s (%cd)||%C(auto)%d%n' --date=relative --color",
                    path=repo_path,
                )

                commit_msg, branch_status = commit.strip().split("||")
                unstaged_symbol = "*" if unstaged else " "
                staged_symbol = "+" if staged else " "
                untracked_symbol = "?" if untracked else " "

                yield [
                    (repo_name, ""),
                    (
                        "Branch",
                        f"{head} {unstaged_symbol}{staged_symbol}{untracked_symbol}",
                    ),
                    ("Status", branch_status),
                    ("Commit Hash", commit_hash),
                    ("Commit Msg", commit_msg),
                    ("Local Path", repo_path),
                ]

    def process_repo_option(self, repos: Optional[List[str]], cmd: str):
        exist_repos = self.load_repos()
        print(f":: {cmd}\n")

        if repos:
            exist_repos = {k: v for k, v in exist_repos.items() if k in repos}

        if len(exist_repos) >= 1:
            return exec_async_tasks(
                async_run_cmd(*cmd.split(), cwd=prop["path"], msg=f":: {prop['path']}")
                for name, prop in exist_repos.items()
            )

        for _, prop in exist_repos.items():
            exec_cmd(cmd, prop["path"])
